@@ -23,21 +23,31 @@
 #include <setjmp.h>
 #endif
 
-#ifdef HAVE_LINUX_FS_H
-#include <linux/fs.h>
-#endif
-
 #if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
 #include "ext2fs.h"
 #else
-#include <linux/ext2_fs.h>
+#include "ext2fs/ext2_fs.h"
 #include "ext2fs/ext2fs.h"
 #endif
 
-
-/* Everything is STDC, these days */
-#define NOARGS void
+#ifdef ENABLE_NLS
+#include <libintl.h>
+#include <locale.h>
+#define _(a) (gettext (a))
+#ifdef gettext_noop
+#define N_(a) gettext_noop (a)
+#else
+#define N_(a) (a)
+#endif
+/* FIXME */
+#define NLS_CAT_NAME "e2fsprogs"
+#define LOCALEDIR "/usr/share/locale"
+/* FIXME */
+#else
+#define _(a) (a)
+#define N_(a) a
+#endif
 
 /*
  * Exit codes used by fsck-type programs
@@ -62,9 +72,9 @@
  * directory information.
  */
 struct dir_info {
-	ino_t			ino;	/* Inode number */
-	ino_t			dotdot;	/* Parent according to '..' */
-	ino_t			parent; /* Parent according to treewalk */
+	ext2_ino_t		ino;	/* Inode number */
+	ext2_ino_t		dotdot;	/* Parent according to '..' */
+	ext2_ino_t		parent; /* Parent according to treewalk */
 };
 
 #ifdef RESOURCE_TRACK
@@ -104,6 +114,7 @@ struct resource_track {
 
 #define E2F_FLAG_PROG_BAR	0x0020 /* Progress bar on screen */
 #define E2F_FLAG_PROG_SUPPRESS	0x0040 /* Progress suspended */
+#define E2F_FLAG_JOURNAL_INODE	0x0080 /* Create a new ext3 journal inode */
 
 /*
  * Defines for indicating the e2fsck pass number
@@ -148,7 +159,6 @@ struct e2fsck_struct {
 
 	ext2fs_block_bitmap block_found_map; /* Blocks which are in use */
 	ext2fs_block_bitmap block_dup_map; /* Blks referenced more than once */
-	ext2fs_block_bitmap block_illegal_map; /* Meta-data blocks */
 
 	/*
 	 * Inode count arrays
@@ -173,7 +183,7 @@ struct e2fsck_struct {
 	/*
 	 * For pass1_check_directory and pass1_get_blocks
 	 */
-	ino_t stashed_ino;
+	ext2_ino_t stashed_ino;
 	struct ext2_inode *stashed_inode;
 
 	/*
@@ -202,6 +212,7 @@ struct e2fsck_struct {
 	int progress_fd;
 	int progress_pos;
 	int progress_last_percent;
+	unsigned int progress_last_time;
 
 	/* File counts */
 	int fs_directory_count;
@@ -253,8 +264,8 @@ extern void read_bad_blocks_file(e2fsck_t ctx, const char *bad_blocks_file,
 extern void test_disk(e2fsck_t ctx);
 
 /* dirinfo.c */
-extern void e2fsck_add_dir_info(e2fsck_t ctx, ino_t ino, ino_t parent);
-extern struct dir_info *e2fsck_get_dir_info(e2fsck_t ctx, ino_t ino);
+extern void e2fsck_add_dir_info(e2fsck_t ctx, ext2_ino_t ino, ext2_ino_t parent);
+extern struct dir_info *e2fsck_get_dir_info(e2fsck_t ctx, ext2_ino_t ino);
 extern void e2fsck_free_dir_info(e2fsck_t ctx);
 extern int e2fsck_get_num_dirs(e2fsck_t ctx);
 extern int e2fsck_get_num_dirinfo(e2fsck_t ctx);
@@ -264,15 +275,19 @@ extern struct dir_info *e2fsck_dir_info_iter(e2fsck_t ctx, int *control);
 extern const char *ehandler_operation(const char *op);
 extern void ehandler_init(io_channel channel);
 
+/* journal.c */
+extern int e2fsck_check_ext3_journal(e2fsck_t ctx);
+extern int e2fsck_run_ext3_journal(e2fsck_t ctx);
+
 /* pass1.c */
 extern void e2fsck_use_inode_shortcuts(e2fsck_t ctx, int bool);
 extern int e2fsck_pass1_check_device_inode(struct ext2_inode *inode);
 
 /* pass2.c */
-extern int e2fsck_process_bad_inode(e2fsck_t ctx, ino_t dir, ino_t ino);
+extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir, ext2_ino_t ino);
 
 /* pass3.c */
-extern int e2fsck_reconnect_file(e2fsck_t ctx, ino_t inode);
+extern int e2fsck_reconnect_file(e2fsck_t ctx, ext2_ino_t inode);
 
 /* super.c */
 void check_super_block(e2fsck_t ctx);
