@@ -38,10 +38,10 @@ int invalid_bitmaps = 0;
 
 struct resource_track	global_rtrack;
 
-static void usage(NOARGS)
+static void usage(void)
 {
 	fprintf(stderr,
-		"Usage: %s [-F] [-I inode_buffer_blocks] device\n",
+		_("Usage: %s [-F] [-I inode_buffer_blocks] device\n"),
 		program_name);
 	exit(1);
 }
@@ -63,12 +63,7 @@ static void PRS(int argc, char *argv[])
 	while ((c = getopt (argc, argv, "FI")) != EOF)
 		switch (c) {
 		case 'F':
-#ifdef BLKFLSBUF
 			flush = 1;
-#else
-			fprintf(stderr, "-F not supported");
-			exit(1);
-#endif
 			break;
 		case 'I':
 			inode_buffer_blocks = atoi(optarg);
@@ -78,24 +73,19 @@ static void PRS(int argc, char *argv[])
 		}
 	device_name = argv[optind];
 	if (flush) {
-#ifdef BLKFLSBUF
 		int	fd = open(device_name, O_RDONLY, 0);
 
 		if (fd < 0) {
-			com_err("open", errno, "while opening %s for flushing",
-				device_name);
+			com_err("open", errno,
+			    _("while opening %s for flushing"), device_name);
 			exit(FSCK_ERROR);
 		}
-		if (ioctl(fd, BLKFLSBUF, 0) < 0) {
-			com_err("BLKFLSBUF", errno, "while trying to flush %s",
-				device_name);
+		if ((retval = ext2fs_sync_device(fd, 1))) {
+			com_err("ext2fs_sync_device", retval,
+				_("while trying to flush %s"), device_name);
 			exit(FSCK_ERROR);
 		}
 		close(fd);
-#else
-		fprintf(stderr, "BLKFLSBUF not supported");
-		exit(1);
-#endif /* BLKFLSBUF */
 	}
 }
 					
@@ -104,7 +94,7 @@ int main (int argc, char *argv[])
 	errcode_t	retval = 0;
 	int		exit_value = FSCK_OK;
 	ext2_filsys	fs;
-	ino_t	ino;
+	ext2_ino_t	ino;
 	int	num_inodes = 0;
 	struct ext2_inode inode;
 	ext2_inode_scan	scan;
@@ -116,7 +106,7 @@ int main (int argc, char *argv[])
 	retval = ext2fs_open(device_name, 0,
 			     0, 0, unix_io_manager, &fs);
 	if (retval) {
-		com_err(program_name, retval, "while trying to open %s",
+		com_err(program_name, retval, _("while trying to open %s"),
 			device_name);
 		exit(1);
 	}
@@ -125,7 +115,7 @@ int main (int argc, char *argv[])
 	
 	retval = ext2fs_open_inode_scan(fs, inode_buffer_blocks, &scan);
 	if (retval) {
-		com_err(program_name, retval, "while opening inode scan");
+		com_err(program_name, retval, _("while opening inode scan"));
 		exit(1);
 	}
 
@@ -133,7 +123,7 @@ int main (int argc, char *argv[])
 		retval = ext2fs_get_next_inode(scan, &ino, &inode);
 		if (retval) {
 			com_err(program_name, retval,
-				"while getting next inode");
+				_("while getting next inode"));
 			exit(1);
 		}
 		if (ino == 0)
@@ -142,7 +132,7 @@ int main (int argc, char *argv[])
 	}
 	
 	print_resource_track(NULL, &global_rtrack);
-	printf("%d inodes scanned.\n", num_inodes);
+	printf(_("%d inodes scanned.\n"), num_inodes);
 	
 	exit(0);
 }
