@@ -32,6 +32,8 @@
  * 	%IF	<inode> -> i_faddr
  * 	%If	<inode> -> i_file_acl
  * 	%Id	<inode> -> i_dir_acl
+ * 	%Iu	<inode> -> i_uid
+ * 	%Ig	<inode> -> i_gid
  * 	%j	<ino2>			inode number
  * 	%m	<com_err error message>
  * 	%N	<num>
@@ -44,15 +46,15 @@
  * 			the containing directory.
  * 	%s	<str>			miscellaneous string
  * 	%S	backup superblock
+ * 	%X	<num> hexadecimal format
  *
  * The following '@' expansions are supported:
  *
  * 	@A	error allocating
  * 	@b	block
  * 	@B	bitmap
+ * 	@c	compress
  * 	@C	conflicts with some other fs block
- * 	@i	inode
- * 	@I	illegal
  * 	@D	deleted
  * 	@d	directory
  * 	@e	entry
@@ -60,12 +62,17 @@
  * 	@f	filesystem
  * 	@F	for @i %i (%Q) is
  * 	@g	group
+ * 	@i	inode
+ * 	@I	illegal
+ * 	@j	journal
  * 	@l	lost+found
  * 	@L	is a link
- * 	@u	unattached
+ * 	@o	orphaned
  * 	@r	root inode
  * 	@s	should be 
  * 	@S	superblock
+ * 	@u	unattached
+ * 	@v	device
  * 	@z	zero-length
  */
 
@@ -92,26 +99,30 @@
  * letter <i> in the table below.
  */
 static const char *abbrevs[] = {
-	"Aerror allocating",
-	"bblock",
-	"Bbitmap",
-	"Cconflicts with some other fs @b",
-	"iinode",
-	"Iillegal",
-	"Ddeleted",
-	"ddirectory",
-	"eentry",
-	"E@e '%Dn' in %p (%i)",
-	"ffilesystem",
-	"Ffor @i %i (%Q) is",
-	"ggroup",
-	"llost+found",
-	"Lis a link",
-	"uunattached",
-	"rroot @i",
-	"sshould be",
-	"Ssuper@b",
-	"zzero-length",
+	N_("Aerror allocating"),
+	N_("bblock"),
+	N_("Bbitmap"),
+	N_("ccompress"),
+	N_("Cconflicts with some other fs @b"),
+	N_("iinode"),
+	N_("Iillegal"),
+	N_("jjournal"),
+	N_("Ddeleted"),
+	N_("ddirectory"),
+	N_("eentry"),
+	N_("E@e '%Dn' in %p (%i)"),
+	N_("ffilesystem"),
+	N_("Ffor @i %i (%Q) is"),
+	N_("ggroup"),
+	N_("llost+found"),
+	N_("Lis a link"),
+	N_("oorphaned"),
+	N_("rroot @i"),
+	N_("sshould be"),
+	N_("Ssuper@b"),
+	N_("uunattached"),
+	N_("vdevice"),
+	N_("zzero-length"),
 	"@@",
 	0
 	};
@@ -119,16 +130,20 @@ static const char *abbrevs[] = {
 /*
  * Give more user friendly names to the "special" inodes.
  */
-#define num_special_inodes	7
+#define num_special_inodes	11
 static const char *special_inode_name[] =
 {
-	"<The NULL inode>",			/* 0 */
-	"<The bad blocks inode>", 		/* 1 */
+	N_("<The NULL inode>"),			/* 0 */
+	N_("<The bad blocks inode>"),		/* 1 */
 	"/",					/* 2 */
-	"<The ACL index inode>",		/* 3 */
-	"<The ACL data inode>",			/* 4 */
-	"<The boot loader inode>",		/* 5 */
-	"<The undelete directory inode>"	/* 6 */
+	N_("<The ACL index inode>"),		/* 3 */
+	N_("<The ACL data inode>"),		/* 4 */
+	N_("<The boot loader inode>"),		/* 5 */
+	N_("<The undelete directory inode>"),	/* 6 */
+	N_("<The group descriptor inode>"),	/* 7 */
+	N_("<The journal inode>"),		/* 8 */
+	N_("<Reserved inode 9>"),		/* 9 */
+	N_("<Reserved inode 10>"),		/* 10 */
 };
 
 /*
@@ -148,9 +163,9 @@ static void safe_print(const char *cp, int len)
 			fputs("M-", stdout);
 			ch -= 128;
 		}
-		if (ch < 32) {
+		if ((ch < 32) || (ch == 0x7f)) {
 			fputc('^', stdout);
-			ch += 32;
+			ch ^= 0x40; /* ^@, ^A, ^B; ^? for DEL */
 		}
 		fputc(ch, stdout);
 	}
@@ -161,13 +176,13 @@ static void safe_print(const char *cp, int len)
  * This function prints a pathname, using the ext2fs_get_pathname
  * function
  */
-static void print_pathname(ext2_filsys fs, ino_t dir, ino_t ino)
+static void print_pathname(ext2_filsys fs, ext2_ino_t dir, ext2_ino_t ino)
 {
 	errcode_t	retval;
 	char		*path;
 
 	if (!dir && (ino < num_special_inodes)) {
-		fputs(special_inode_name[ino], stdout);
+		fputs(_(special_inode_name[ino]), stdout);
 		return;
 	}
 	
@@ -202,13 +217,13 @@ static _INLINE_ void expand_at_expression(e2fsck_t ctx, char ch,
 			*first = 0;
 			fputc(toupper(*str++), stdout);
 		}
-		print_e2fsck_message(ctx, str, pctx, *first);
+		print_e2fsck_message(ctx, _(str), pctx, *first);
 	} else
 		printf("@%c", ch);
 }
 
 /*
- * This function expands '%kX' expressions
+ * This function expands '%IX' expressions
  */
 static _INLINE_ void expand_inode_expression(char ch,
 					       struct problem_context *ctx)
@@ -263,6 +278,14 @@ static _INLINE_ void expand_inode_expression(char ch,
 		printf("%u", (LINUX_S_ISDIR(inode->i_mode) ?
 			      inode->i_dir_acl : 0));
 		break;
+	case 'u':
+		printf("%d", (inode->i_uid |
+			      (inode->osd2.linux2.l_i_uid_high << 16)));
+		break;
+	case 'g':
+		printf("%d", (inode->i_gid |
+			      (inode->osd2.linux2.l_i_gid_high << 16)));
+		break;
 	default:
 	no_inode:
 		printf("%%I%c", ch);
@@ -336,16 +359,16 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 		printf("%u", ctx->blk2);
 		break;
 	case 'd':
-		printf("%lu", ctx->dir);
+		printf("%u", ctx->dir);
 		break;
 	case 'g':
 		printf("%d", ctx->group);
 		break;
 	case 'i':
-		printf("%lu", ctx->ino);
+		printf("%u", ctx->ino);
 		break;
 	case 'j':
-		printf("%lu", ctx->ino2);
+		printf("%u", ctx->ino2);
 		break;
 	case 'm':
 		printf("%s", error_message(ctx->errcode));
@@ -374,7 +397,14 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 		printf("%d", get_backup_sb(fs));
 		break;
 	case 's':
-		printf("%s", ctx->str);
+		printf("%s", ctx->str ? ctx->str : "NULL");
+		break;
+	case 'X':
+#ifdef EXT2_NO_64_TYPE
+		printf("0x%x", ctx->num);
+#else
+		printf("0x%llx", ctx->num);
+#endif
 		break;
 	default:
 	no_context:
