@@ -60,9 +60,9 @@ static int process_bad_block(ext2_filsys fs, blk_t *block_nr,
 static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 			 char *block_buf);
 static void mark_table_blocks(e2fsck_t ctx);
-static void alloc_bad_map(e2fsck_t ctx);
 static void alloc_bb_map(e2fsck_t ctx);
 static void alloc_imagic_map(e2fsck_t ctx);
+static void mark_inode_bad(e2fsck_t ctx, ino_t ino);
 static void handle_fs_bad_blocks(e2fsck_t ctx);
 static void process_inodes(e2fsck_t ctx, char *block_buf);
 static EXT2_QSORT_TYPE process_inode_cmp(const void *a, const void *b);
@@ -71,8 +71,9 @@ static errcode_t scan_callback(ext2_filsys fs, ext2_inode_scan scan,
 /* static char *describe_illegal_block(ext2_filsys fs, blk_t block); */
 
 struct process_block_struct {
-	ino_t		ino;
-	int		is_dir:1, clear:1, suppress:1, fragmented:1;
+	ext2_ino_t	ino;
+	int		is_dir:1, clear:1, suppress:1,
+				fragmented:1, compressed:1;
 	blk_t		num_blocks;
 	e2_blkcnt_t	last_block;
 	int		num_illegal_blocks;
@@ -83,7 +84,7 @@ struct process_block_struct {
 };
 
 struct process_inode_block {
-	ino_t	ino;
+	ext2_ino_t ino;
 	struct ext2_inode inode;
 };
 
@@ -133,7 +134,7 @@ int e2fsck_pass1_check_device_inode(struct ext2_inode *inode)
 	 * you can't set or clear immutable flags for devices.)  Once
 	 * the kernel has been fixed we can change this...
 	 */
-	if (inode->i_flags & EXT2_IMMUTABLE_FL) {
+	if (inode->i_flags & (EXT2_IMMUTABLE_FL | EXT2_APPEND_FL)) {
 		for (i=4; i < EXT2_N_BLOCKS; i++) 
 			if (inode->i_block[i])
 				return 0;
@@ -142,27 +143,51 @@ int e2fsck_pass1_check_device_inode(struct ext2_inode *inode)
 }
 
 /*
- * If the immutable flag is set on the inode, offer to clear it.
+ * If the immutable (or append-only) flag is set on the inode, offer
+ * to clear it.
  */
 static void check_immutable(e2fsck_t ctx, struct problem_context *pctx)
 {
-	if (!(pctx->inode->i_flags & EXT2_IMMUTABLE_FL))
+	if (!(pctx->inode->i_flags & (EXT2_IMMUTABLE_FL | EXT2_APPEND_FL)))
 		return;
 
 	if (!fix_problem(ctx, PR_1_SET_IMMUTABLE, pctx))
 		return;
 
-	pctx->inode->i_flags &= ~EXT2_IMMUTABLE_FL;
+	pctx->inode->i_flags &= ~((EXT2_IMMUTABLE_FL | EXT2_APPEND_FL));
 	e2fsck_write_inode(ctx, pctx->ino, pctx->inode, "pass1");
 }
 
+/*
+ * If device, fifo or socket, check size is zero -- if not offer to
+ * clear it
+ */
+static void check_size(e2fsck_t ctx, struct problem_context *pctx)
+{
+	struct ext2_inode *inode = pctx->inode;
+	
+	if ((LINUX_S_ISBLK(inode->i_mode) ||
+	     LINUX_S_ISCHR(inode->i_mode) ||
+	     LINUX_S_ISFIFO(inode->i_mode) ||
+	     LINUX_S_ISSOCK(inode->i_mode)) && 
+	    !inode->i_size || !inode->i_size_high)
+		return;
+	
+	if(!fix_problem(ctx, PR_1_SET_NONZSIZE, pctx))
+		return;
+	
+	inode->i_size = 0;
+	inode->i_size_high = 0;
+	e2fsck_write_inode(ctx, pctx->ino, pctx->inode, "pass1");
+}
+	
 
 void e2fsck_pass1(e2fsck_t ctx)
 {
 	int	i;
-	__u64	max_sizes;
+	__u64	max_sizes, max_sect_limit;
 	ext2_filsys fs = ctx->fs;
-	ino_t	ino;
+	ext2_ino_t	ino;
 	struct ext2_inode inode;
 	ext2_inode_scan	scan;
 	char		*block_buf;
@@ -172,7 +197,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 	unsigned char	frag, fsize;
 	struct		problem_context pctx;
 	struct		scan_callback_struct scan_struct;
-	struct ext2fs_sb *sb;
+	struct ext2_super_block *sb = ctx->fs->super;
 	int		imagic_fs;
 	
 #ifdef RESOURCE_TRACK
@@ -196,17 +221,19 @@ void e2fsck_pass1(e2fsck_t ctx)
 			     (__u64) EXT2_BPP(10+i) * EXT2_BPP(10+i) *
 			     EXT2_BPP(10+i));
 		max_sizes = (max_sizes * (1UL << (10+i))) - 1;
+		max_sect_limit = 512ULL * ((1LL << 32) - (1 << (i+1)));
+		if (max_sizes > max_sect_limit)
+			max_sizes = max_sect_limit;
 		ext2_max_sizes[i] = max_sizes;
 	}
 #undef EXT2_BPP
 
-	sb = (struct ext2fs_sb *) fs->super;
 	imagic_fs = (sb->s_feature_compat & EXT2_FEATURE_COMPAT_IMAGIC_INODES);
 
 	/*
 	 * Allocate bitmaps structures
 	 */
-	pctx.errcode = ext2fs_allocate_inode_bitmap(fs, "in-use inode map",
+	pctx.errcode = ext2fs_allocate_inode_bitmap(fs, _("in-use inode map"),
 					      &ctx->inode_used_map);
 	if (pctx.errcode) {
 		pctx.num = 1;
@@ -214,8 +241,8 @@ void e2fsck_pass1(e2fsck_t ctx)
 		ctx->flags |= E2F_FLAG_ABORT;
 		return;
 	}
-	pctx.errcode = ext2fs_allocate_inode_bitmap(fs, "directory inode map",
-					      &ctx->inode_dir_map);
+	pctx.errcode = ext2fs_allocate_inode_bitmap(fs,
+				_("directory inode map"), &ctx->inode_dir_map);
 	if (pctx.errcode) {
 		pctx.num = 2;
 		fix_problem(ctx, PR_1_ALLOCATE_IBITMAP_ERROR, &pctx);
@@ -223,15 +250,14 @@ void e2fsck_pass1(e2fsck_t ctx)
 		return;
 	}
 	pctx.errcode = ext2fs_allocate_inode_bitmap(fs,
-						    "regular file inode map",
-					      &ctx->inode_reg_map);
+			_("regular file inode map"), &ctx->inode_reg_map);
 	if (pctx.errcode) {
 		pctx.num = 6;
 		fix_problem(ctx, PR_1_ALLOCATE_IBITMAP_ERROR, &pctx);
 		ctx->flags |= E2F_FLAG_ABORT;
 		return;
 	}
-	pctx.errcode = ext2fs_allocate_block_bitmap(fs, "in-use block map",
+	pctx.errcode = ext2fs_allocate_block_bitmap(fs, _("in-use block map"),
 					      &ctx->block_found_map);
 	if (pctx.errcode) {
 		pctx.num = 1;
@@ -239,14 +265,6 @@ void e2fsck_pass1(e2fsck_t ctx)
 		ctx->flags |= E2F_FLAG_ABORT;
 		return;
 	}
-	pctx.errcode = ext2fs_allocate_block_bitmap(fs, "illegal block map",
-					      &ctx->block_illegal_map);
-	if (pctx.errcode) {
-		pctx.num = 2;
-		fix_problem(ctx, PR_1_ALLOCATE_BBITMAP_ERROR, &pctx);
-		ctx->flags |= E2F_FLAG_ABORT;
-		return;
-	}
 	pctx.errcode = ext2fs_create_icount2(fs, 0, 0, 0,
 					     &ctx->inode_link_info);
 	if (pctx.errcode) {
@@ -268,11 +286,25 @@ void e2fsck_pass1(e2fsck_t ctx)
 		return;
 	}
 
+	/*
+	 * If the last orphan field is set, clear it, since the pass1
+	 * processing will automatically find and clear the orphans.
+	 * In the future, we may want to try using the last_orphan
+	 * linked list ourselves, but for now, we clear it so that the
+	 * ext3 mount code won't get confused.
+	 */
+	if (!(ctx->options & E2F_OPT_READONLY)) {
+		if (fs->super->s_last_orphan) {
+			fs->super->s_last_orphan = 0;
+			ext2fs_mark_super_dirty(fs);
+		}
+	}
+
 	mark_table_blocks(ctx);
 	block_buf = (char *) e2fsck_allocate_memory(ctx, fs->blocksize * 3,
 						    "block interate buffer");
 	e2fsck_use_inode_shortcuts(ctx, 1);
-	ehandler_operation("doing inode scan");
+	ehandler_operation(_("doing inode scan"));
 	pctx.errcode = ext2fs_open_inode_scan(fs, ctx->inode_buffer_blocks, 
 					      &scan);
 	if (pctx.errcode) {
@@ -329,8 +361,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 			ext2fs_mark_inode_bitmap(ctx->inode_used_map, ino);
 			clear_problem_context(&pctx);
 			goto next;
-		}
-		if (ino == EXT2_ROOT_INO) {
+		} else if (ino == EXT2_ROOT_INO) {
 			/*
 			 * Make sure the root inode is a directory; if
 			 * not, offer to clear it.  It will be
@@ -345,6 +376,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 					e2fsck_write_inode(ctx, ino, &inode,
 							   "pass1");
 				}
+
 			}
 			/*
 			 * If dtime is set, offer to clear it.  mke2fs
@@ -362,16 +394,46 @@ void e2fsck_pass1(e2fsck_t ctx)
 							   "pass1");
 				}
 			}
-		}
-		if ((ino != EXT2_ROOT_INO) &&
-		    (ino < EXT2_FIRST_INODE(fs->super))) {
+		} else if (ino == EXT2_JOURNAL_INO) {
 			ext2fs_mark_inode_bitmap(ctx->inode_used_map, ino);
-			if (((ino == EXT2_BOOT_LOADER_INO) &&
-			     LINUX_S_ISDIR(inode.i_mode)) ||
-			    ((ino != EXT2_BOOT_LOADER_INO) &&
-			     (inode.i_mode != 0))) {
-				if (fix_problem(ctx,
-					    PR_1_RESERVED_BAD_MODE, &pctx)) {
+			if (fs->super->s_journal_inum == EXT2_JOURNAL_INO) {
+				/*
+				 * XXX arguably this check should be
+				 * in journal.c, before we decide it's
+				 * safe to run the journal...
+				 */
+				if (!LINUX_S_ISREG(inode.i_mode) &&
+				    fix_problem(ctx, PR_1_JOURNAL_BAD_MODE,
+						&pctx)) {
+					inode.i_mode = LINUX_S_IFREG;
+					e2fsck_write_inode(ctx, ino, &inode,
+							   "pass1");
+				}
+				check_blocks(ctx, &pctx, block_buf);
+				goto next;
+			}
+			if ((inode.i_links_count || inode.i_blocks ||
+			     inode.i_blocks || inode.i_block[0]) &&
+			    fix_problem(ctx, PR_1_JOURNAL_INODE_NOT_CLEAR, 
+					&pctx)) {
+				memset(&inode, 0, sizeof(inode));
+				ext2fs_icount_store(ctx->inode_link_info,
+						    ino, 0);
+				e2fsck_write_inode(ctx, ino, &inode, "pass1");
+			}
+		} else if (ino < EXT2_FIRST_INODE(fs->super)) {
+			int	problem = 0;
+			
+			ext2fs_mark_inode_bitmap(ctx->inode_used_map, ino);
+			if (ino == EXT2_BOOT_LOADER_INO) {
+				if (LINUX_S_ISDIR(inode.i_mode))
+					problem = PR_1_RESERVED_BAD_MODE;
+			} else {
+				if (inode.i_mode != 0)
+					problem = PR_1_RESERVED_BAD_MODE;
+			}
+			if (problem) {
+				if (fix_problem(ctx, problem, &pctx)) {
 					inode.i_mode = 0;
 					e2fsck_write_inode(ctx, ino, &inode,
 							   "pass1");
@@ -380,6 +442,33 @@ void e2fsck_pass1(e2fsck_t ctx)
 			check_blocks(ctx, &pctx, block_buf);
 			goto next;
 		}
+		/*
+		 * Check for inodes who might have been part of the
+		 * orphaned list linked list.  They should have gotten
+		 * dealt with by now, unless the list had somehow been
+		 * corrupted.
+		 * 
+		 * FIXME: In the future, inodes which are still in use
+		 * (and which are therefore) pending truncation should
+		 * be handled specially.  Right now we just clear the
+		 * dtime field, and the normal e2fsck handling of
+		 * inodes where i_size and the inode blocks are
+		 * inconsistent is to fix i_size, instead of releasing
+		 * the extra blocks.  This won't catch the inodes that
+		 * was at the end of the orphan list, but it's better
+		 * than nothing.  The right answer is that there
+		 * shouldn't be any bugs in the orphan list handling.  :-)
+		 */
+		if (inode.i_dtime &&
+		    inode.i_dtime < ctx->fs->super->s_inodes_count) {
+			if (fix_problem(ctx, PR_1_LOW_DTIME, &pctx)) {
+				inode.i_dtime = inode.i_links_count ?
+					0 : time(0);
+				e2fsck_write_inode(ctx, ino, &inode,
+						   "pass1");
+			}
+		}
+		
 		/*
 		 * This code assumes that deleted inodes have
 		 * i_links_count set to 0.  
@@ -432,11 +521,8 @@ void e2fsck_pass1(e2fsck_t ctx)
 		
 		if (inode.i_faddr || frag || fsize
 		    || inode.i_file_acl ||
-		    (LINUX_S_ISDIR(inode.i_mode) && inode.i_dir_acl)) {
-			if (!ctx->inode_bad_map)
-				alloc_bad_map(ctx);
-			ext2fs_mark_inode_bitmap(ctx->inode_bad_map, ino);
-		}
+		    (LINUX_S_ISDIR(inode.i_mode) && inode.i_dir_acl))
+			mark_inode_bad(ctx, ino);
 		if (inode.i_flags & EXT2_IMAGIC_FL) {
 			if (imagic_fs) {
 				if (!ctx->inode_imagic_map)
@@ -462,14 +548,19 @@ void e2fsck_pass1(e2fsck_t ctx)
 		} else if (LINUX_S_ISCHR (inode.i_mode) &&
 			   e2fsck_pass1_check_device_inode(&inode)) {
 			check_immutable(ctx, &pctx);
+			check_size(ctx, &pctx);
 			ctx->fs_chardev_count++;
 		} else if (LINUX_S_ISBLK (inode.i_mode) &&
 			   e2fsck_pass1_check_device_inode(&inode)) {
 			check_immutable(ctx, &pctx);
+			check_size(ctx, &pctx);
 			ctx->fs_blockdev_count++;
 		} else if (LINUX_S_ISLNK (inode.i_mode)) {
 			ctx->fs_symlinks_count++;
 			if (!inode.i_blocks) {
+				if (inode.i_size_high ||
+				    (inode.i_size > EXT2_N_BLOCKS*4))
+					mark_inode_bad(ctx, ino);
 				ctx->fs_fast_symlinks_count++;
 				goto next;
 			}
@@ -477,16 +568,15 @@ void e2fsck_pass1(e2fsck_t ctx)
 		else if (LINUX_S_ISFIFO (inode.i_mode) &&
 			 e2fsck_pass1_check_device_inode(&inode)) {
 			check_immutable(ctx, &pctx);
+			check_size(ctx, &pctx);
 			ctx->fs_fifo_count++;
 		} else if ((LINUX_S_ISSOCK (inode.i_mode)) &&
 			   e2fsck_pass1_check_device_inode(&inode)) {
 			check_immutable(ctx, &pctx);
-		        ctx->fs_sockets_count++;
-		} else {
-			if (!ctx->inode_bad_map)
-				alloc_bad_map(ctx);
-			ext2fs_mark_inode_bitmap(ctx->inode_bad_map, ino);
-		}
+			check_size(ctx, &pctx);
+			ctx->fs_sockets_count++;
+		} else
+			mark_inode_bad(ctx, ino);
 		if (inode.i_block[EXT2_IND_BLOCK])
 			ctx->fs_ind_count++;
 		if (inode.i_block[EXT2_DIND_BLOCK])
@@ -559,17 +649,20 @@ endit:
 	e2fsck_use_inode_shortcuts(ctx, 0);
 	
 	ext2fs_free_mem((void **) &block_buf);
-	ext2fs_free_block_bitmap(ctx->block_illegal_map);
-	ctx->block_illegal_map = 0;
 
-	if (ctx->large_files && 
-	    !(sb->s_feature_ro_compat & 
-	      EXT2_FEATURE_RO_COMPAT_LARGE_FILE)) {
-		if (fix_problem(ctx, PR_1_FEATURE_LARGE_FILES, &pctx)) {
-			sb->s_feature_ro_compat |= 
+	if (ctx->large_files) {
+		if (!(sb->s_feature_ro_compat &
+		      EXT2_FEATURE_RO_COMPAT_LARGE_FILE) &&
+		    fix_problem(ctx, PR_1_FEATURE_LARGE_FILES, &pctx)) {
+			sb->s_feature_ro_compat |=
 				EXT2_FEATURE_RO_COMPAT_LARGE_FILE;
 			ext2fs_mark_super_dirty(fs);
 		}
+		if (sb->s_rev_level == EXT2_GOOD_OLD_REV &&
+		    fix_problem(ctx, PR_1_FS_REV_LEVEL, &pctx)) {
+			ext2fs_update_dynamic_rev(fs);
+			ext2fs_mark_super_dirty(fs);
+		}
 	} else if (!ctx->large_files &&
 	    (sb->s_feature_ro_compat &
 	      EXT2_FEATURE_RO_COMPAT_LARGE_FILE)) {
@@ -583,7 +676,7 @@ endit:
 #ifdef RESOURCE_TRACK
 	if (ctx->options & E2F_OPT_TIME2) {
 		e2fsck_clear_progbar(ctx);
-		print_resource_track("Pass 1", &rtrack);
+		print_resource_track(_("Pass 1"), &rtrack);
 	}
 #endif
 }
@@ -618,7 +711,7 @@ static void process_inodes(e2fsck_t ctx, char *block_buf)
 {
 	int			i;
 	struct ext2_inode	*old_stashed_inode;
-	ino_t			old_stashed_ino;
+	ext2_ino_t		old_stashed_ino;
 	const char		*old_operation;
 	char			buf[80];
 	struct problem_context	pctx;
@@ -626,6 +719,8 @@ static void process_inodes(e2fsck_t ctx, char *block_buf)
 #if 0
 	printf("begin process_inodes: ");
 #endif
+	if (process_inode_count == 0)
+		return;
 	old_operation = ehandler_operation(0);
 	old_stashed_inode = ctx->stashed_inode;
 	old_stashed_ino = ctx->stashed_ino;
@@ -639,7 +734,8 @@ static void process_inodes(e2fsck_t ctx, char *block_buf)
 #if 0
 		printf("%u ", pctx.ino);
 #endif
-		sprintf(buf, "reading indirect blocks of inode %lu", pctx.ino);
+		sprintf(buf, _("reading indirect blocks of inode %u"),
+			pctx.ino);
 		ehandler_operation(buf);
 		check_blocks(ctx, &pctx, block_buf);
 		if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
@@ -666,25 +762,29 @@ static EXT2_QSORT_TYPE process_inode_cmp(const void *a, const void *b)
 }
 
 /*
- * This procedure will allocate the inode bad map table
+ * Mark an inode as being bad in some what
  */
-static void alloc_bad_map(e2fsck_t ctx)
+static void mark_inode_bad(e2fsck_t ctx, ino_t ino)
 {
 	struct		problem_context pctx;
+
+	if (!ctx->inode_bad_map) {
+		clear_problem_context(&pctx);
 	
-	clear_problem_context(&pctx);
-	
-	pctx.errcode = ext2fs_allocate_inode_bitmap(ctx->fs, "bad inode map",
-					      &ctx->inode_bad_map);
-	if (pctx.errcode) {
-		pctx.num = 3;
-		fix_problem(ctx, PR_1_ALLOCATE_IBITMAP_ERROR, &pctx);
-		/* Should never get here */
-		ctx->flags |= E2F_FLAG_ABORT;
-		return;
+		pctx.errcode = ext2fs_allocate_inode_bitmap(ctx->fs,
+			    _("bad inode map"), &ctx->inode_bad_map);
+		if (pctx.errcode) {
+			pctx.num = 3;
+			fix_problem(ctx, PR_1_ALLOCATE_IBITMAP_ERROR, &pctx);
+			/* Should never get here */
+			ctx->flags |= E2F_FLAG_ABORT;
+			return;
+		}
 	}
+	ext2fs_mark_inode_bitmap(ctx->inode_bad_map, ino);
 }
 
+
 /*
  * This procedure will allocate the inode "bb" (badblock) map table
  */
@@ -694,7 +794,7 @@ static void alloc_bb_map(e2fsck_t ctx)
 	
 	clear_problem_context(&pctx);
 	pctx.errcode = ext2fs_allocate_inode_bitmap(ctx->fs,
-					      "inode in bad block map",
+					      _("inode in bad block map"),
 					      &ctx->inode_bb_map);
 	if (pctx.errcode) {
 		pctx.num = 4;
@@ -714,7 +814,7 @@ static void alloc_imagic_map(e2fsck_t ctx)
 	
 	clear_problem_context(&pctx);
 	pctx.errcode = ext2fs_allocate_inode_bitmap(ctx->fs,
-					      "imagic inode map",
+					      _("imagic inode map"),
 					      &ctx->inode_imagic_map);
 	if (pctx.errcode) {
 		pctx.num = 5;
@@ -741,7 +841,7 @@ static _INLINE_ void mark_block_used(e2fsck_t ctx, blk_t block)
 	if (ext2fs_fast_test_block_bitmap(ctx->block_found_map, block)) {
 		if (!ctx->block_dup_map) {
 			pctx.errcode = ext2fs_allocate_block_bitmap(ctx->fs,
-			      "multiply claimed block map",
+			      _("multiply claimed block map"),
 			      &ctx->block_dup_map);
 			if (pctx.errcode) {
 				pctx.num = 3;
@@ -767,11 +867,10 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 {
 	ext2_filsys fs = ctx->fs;
 	struct process_block_struct pb;
-	ino_t		ino = pctx->ino;
+	ext2_ino_t	ino = pctx->ino;
 	struct ext2_inode *inode = pctx->inode;
 	int		bad_size = 0;
 	__u64		size;
-	struct ext2fs_sb	*sb;
 	
 	if (!ext2fs_inode_has_valid_blocks(pctx->inode))
 		return;
@@ -781,12 +880,27 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 	pb.num_illegal_blocks = 0;
 	pb.suppress = 0; pb.clear = 0;
 	pb.fragmented = 0;
+	pb.compressed = 0;
 	pb.previous_block = 0;
 	pb.is_dir = LINUX_S_ISDIR(pctx->inode->i_mode);
 	pb.inode = inode;
 	pb.pctx = pctx;
 	pb.ctx = ctx;
 	pctx->ino = ino;
+
+	if (inode->i_flags & EXT2_COMPRBLK_FL) {
+		if (fs->super->s_feature_incompat &
+		    EXT2_FEATURE_INCOMPAT_COMPRESSION)
+			pb.compressed = 1;
+		else {
+			if (fix_problem(ctx, PR_1_COMPR_SET, pctx)) {
+				inode->i_flags &= ~EXT2_COMPRBLK_FL;
+				e2fsck_write_inode(ctx, ino, inode,
+						   "check_blocks");
+			}
+		}
+	}
+
 	pctx->errcode = ext2fs_block_iterate2(fs, ino,
 				       pb.is_dir ? BLOCK_FLAG_HOLE : 0,
 				       block_buf, process_block, &pb);
@@ -838,21 +952,22 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 	}
 	if (pb.is_dir) {
 		int nblock = inode->i_size >> EXT2_BLOCK_SIZE_BITS(fs->super);
-		if ((nblock > (pb.last_block + 1)) ||
-		    ((inode->i_size & (fs->blocksize-1)) != 0))
+		/* We don't let a directory become larger than 2GB */
+		if (nblock > (pb.last_block + 1) ||
+		    (inode->i_size & ((fs->blocksize-1) | 0x80000000UL)) != 0)
 			bad_size = 1;
 		else if (nblock < (pb.last_block + 1)) {
-			sb = (struct ext2fs_sb *) fs->super;
 			if (((pb.last_block + 1) - nblock) >
-			    sb->s_prealloc_dir_blocks)
+			    fs->super->s_prealloc_dir_blocks)
 				bad_size = 2;
 		}
 	} else {
-		size = inode->i_size + ((__u64) inode->i_size_high << 32);
+		size = inode->i_size | ((__u64) inode->i_size_high << 32);
 		if ((size < pb.last_block * fs->blocksize))
 			bad_size = 3;
 		else if (size > ext2_max_sizes[fs->super->s_log_block_size])
 			bad_size = 4;
+		/* FIXME: need to ensure pb.num_blocks < 2^32 */
 	}
 	if (bad_size) {
 		pctx->num = (pb.last_block+1) * fs->blocksize;
@@ -864,7 +979,7 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 		}
 		pctx->num = 0;
 	}
-	if (!pb.is_dir && inode->i_size_high)
+	if (!pb.is_dir && (inode->i_size_high || inode->i_size & 0x80000000UL))
 		ctx->large_files++;
 	if (pb.num_blocks != inode->i_blocks) {
 		pctx->num = pb.num_blocks;
@@ -931,7 +1046,7 @@ static char *describe_illegal_block(ext2_filsys fs, blk_t block)
 /*
  * This is a helper function for check_blocks().
  */
-int process_block(ext2_filsys fs,
+static int process_block(ext2_filsys fs,
 		  blk_t	*block_nr,
 		  e2_blkcnt_t blockcnt,
 		  blk_t ref_block,
@@ -949,6 +1064,28 @@ int process_block(ext2_filsys fs,
 	pctx = p->pctx;
 	ctx = p->ctx;
 
+	if (p->compressed && (blk == EXT2FS_COMPRESSED_BLKADDR)) {
+		/* todo: Check that the comprblk_fl is high, that the
+		   blkaddr pattern looks right (all non-holes up to
+		   first EXT2FS_COMPRESSED_BLKADDR, then all
+		   EXT2FS_COMPRESSED_BLKADDR up to end of cluster),
+		   that the feature_incompat bit is high, and that the
+		   inode is a regular file.  If we're doing a "full
+		   check" (a concept introduced to e2fsck by e2compr,
+		   meaning that we look at data blocks as well as
+		   metadata) then call some library routine that
+		   checks the compressed data.  I'll have to think
+		   about this, because one particularly important
+		   problem to be able to fix is to recalculate the
+		   cluster size if necessary.  I think that perhaps
+		   we'd better do most/all e2compr-specific checks
+		   separately, after the non-e2compr checks.  If not
+		   doing a full check, it may be useful to test that
+		   the personality is linux; e.g. if it isn't then
+		   perhaps this really is just an illegal block. */
+		return 0;
+	}
+	
 	if (blk == 0) {
 		if (p->is_dir == 0) {
 			/*
@@ -984,7 +1121,7 @@ int process_block(ext2_filsys fs,
 	 * file be contiguous.  (Which can never be true for really
 	 * big files that are greater than a block group.)
 	 */
-	if (p->previous_block) {
+	if (!HOLE_BLKADDR(p->previous_block)) {
 		if (p->previous_block+1 != blk)
 			p->fragmented = 1;
 	}
@@ -993,11 +1130,6 @@ int process_block(ext2_filsys fs,
 	if (blk < fs->super->s_first_data_block ||
 	    blk >= fs->super->s_blocks_count)
 		problem = PR_1_ILLEGAL_BLOCK_NUM;
-#if 0
-	else
-		if (ext2fs_test_block_bitmap(block_illegal_map, blk))
-			problem = PR_1_BLOCK_OVERLAPS_METADATA;
-#endif
 
 	if (problem) {
 		p->num_illegal_blocks++;
@@ -1020,8 +1152,6 @@ int process_block(ext2_filsys fs,
 			goto mark_dir;
 		} else
 			return 0;
-		pctx->blk = 0;
-		pctx->blkcount = -1;
 	}
 
 	mark_block_used(ctx, blk);
@@ -1056,7 +1186,7 @@ static void bad_block_indirect(e2fsck_t ctx, blk_t blk)
 		ctx->flags |= E2F_FLAG_ABORT;
 }
 
-int process_bad_block(ext2_filsys fs,
+static int process_bad_block(ext2_filsys fs,
 		      blk_t *block_nr,
 		      e2_blkcnt_t blockcnt,
 		      blk_t ref_block,
@@ -1070,6 +1200,11 @@ int process_bad_block(ext2_filsys fs,
 	struct problem_context *pctx;
 	e2fsck_t	ctx;
 
+	/*
+	 * Note: This function processes blocks for the bad blocks
+	 * inode, which is never compressed.  So we don't use HOLE_BLKADDR().
+	 */
+
 	if (!blk)
 		return 0;
 	
@@ -1268,15 +1403,15 @@ static void handle_fs_bad_blocks(e2fsck_t ctx)
 
 	for (i = 0; i < fs->group_desc_count; i++) {
 		if (ctx->invalid_block_bitmap_flag[i]) {
-			new_table_block(ctx, first_block, i, "block bitmap", 
+			new_table_block(ctx, first_block, i, _("block bitmap"),
 					1, &fs->group_desc[i].bg_block_bitmap);
 		}
 		if (ctx->invalid_inode_bitmap_flag[i]) {
-			new_table_block(ctx, first_block, i, "inode bitmap", 
+			new_table_block(ctx, first_block, i, _("inode bitmap"),
 					1, &fs->group_desc[i].bg_inode_bitmap);
 		}
 		if (ctx->invalid_inode_table_flag[i]) {
-			new_table_block(ctx, first_block, i, "inode table",
+			new_table_block(ctx, first_block, i, _("inode table"),
 					fs->inode_blocks_per_group, 
 					&fs->group_desc[i].bg_inode_table);
 			ctx->flags |= E2F_FLAG_RESTART;
@@ -1308,8 +1443,6 @@ static void mark_table_blocks(e2fsck_t ctx)
 			 * Mark this group's copy of the superblock
 			 */
 			ext2fs_mark_block_bitmap(ctx->block_found_map, block);
-			ext2fs_mark_block_bitmap(ctx->block_illegal_map,
-						 block);
 		
 			/*
 			 * Mark this group's copy of the descriptors
@@ -1317,8 +1450,6 @@ static void mark_table_blocks(e2fsck_t ctx)
 			for (j = 0; j < fs->desc_blocks; j++) {
 				ext2fs_mark_block_bitmap(ctx->block_found_map,
 							 block + j + 1);
-				ext2fs_mark_block_bitmap(ctx->block_illegal_map,
-							 block + j + 1);
 			}
 		}
 		
@@ -1340,8 +1471,6 @@ static void mark_table_blocks(e2fsck_t ctx)
 				} else {
 				    ext2fs_mark_block_bitmap(ctx->block_found_map,
 							     b);
-				    ext2fs_mark_block_bitmap(ctx->block_illegal_map,
-							     b);
 			    	}
 			}
 		}
@@ -1360,8 +1489,6 @@ static void mark_table_blocks(e2fsck_t ctx)
 			} else {
 			    ext2fs_mark_block_bitmap(ctx->block_found_map,
 				     fs->group_desc[i].bg_block_bitmap);
-			    ext2fs_mark_block_bitmap(ctx->block_illegal_map,
-				     fs->group_desc[i].bg_block_bitmap);
 		    }
 			
 		}
@@ -1379,8 +1506,6 @@ static void mark_table_blocks(e2fsck_t ctx)
 			} else {
 			    ext2fs_mark_block_bitmap(ctx->block_found_map,
 				     fs->group_desc[i].bg_inode_bitmap);
-			    ext2fs_mark_block_bitmap(ctx->block_illegal_map,
-				     fs->group_desc[i].bg_inode_bitmap);
 			}
 		}
 		block += fs->super->s_blocks_per_group;
@@ -1393,12 +1518,13 @@ static void mark_table_blocks(e2fsck_t ctx)
  * structure, so there's no point in letting the ext2fs library read
  * the inode again.
  */
-static errcode_t pass1_get_blocks(ext2_filsys fs, ino_t ino, blk_t *blocks)
+static errcode_t pass1_get_blocks(ext2_filsys fs, ext2_ino_t ino,
+				  blk_t *blocks)
 {
 	e2fsck_t ctx = (e2fsck_t) fs->priv_data;
 	int	i;
 	
-	if (ino != ctx->stashed_ino)
+	if ((ino != ctx->stashed_ino) || !ctx->stashed_inode)
 		return EXT2_ET_CALLBACK_NOTHANDLED;
 
 	for (i=0; i < EXT2_N_BLOCKS; i++)
@@ -1406,32 +1532,32 @@ static errcode_t pass1_get_blocks(ext2_filsys fs, ino_t ino, blk_t *blocks)
 	return 0;
 }
 
-static errcode_t pass1_read_inode(ext2_filsys fs, ino_t ino,
+static errcode_t pass1_read_inode(ext2_filsys fs, ext2_ino_t ino,
 				  struct ext2_inode *inode)
 {
 	e2fsck_t ctx = (e2fsck_t) fs->priv_data;
 
-	if (ino != ctx->stashed_ino)
+	if ((ino != ctx->stashed_ino) || !ctx->stashed_inode)
 		return EXT2_ET_CALLBACK_NOTHANDLED;
 	*inode = *ctx->stashed_inode;
 	return 0;
 }
 
-static errcode_t pass1_write_inode(ext2_filsys fs, ino_t ino,
+static errcode_t pass1_write_inode(ext2_filsys fs, ext2_ino_t ino,
 			    struct ext2_inode *inode)
 {
 	e2fsck_t ctx = (e2fsck_t) fs->priv_data;
 
-	if (ino == ctx->stashed_ino)
+	if ((ino == ctx->stashed_ino) && ctx->stashed_inode)
 		*ctx->stashed_inode = *inode;
 	return EXT2_ET_CALLBACK_NOTHANDLED;
 }
 
-static errcode_t pass1_check_directory(ext2_filsys fs, ino_t ino)
+static errcode_t pass1_check_directory(ext2_filsys fs, ext2_ino_t ino)
 {
 	e2fsck_t ctx = (e2fsck_t) fs->priv_data;
 
-	if (ino != ctx->stashed_ino)
+	if ((ino != ctx->stashed_ino) || !ctx->stashed_inode)
 		return EXT2_ET_CALLBACK_NOTHANDLED;
 
 	if (!LINUX_S_ISDIR(ctx->stashed_inode->i_mode))
@@ -1456,5 +1582,3 @@ void e2fsck_use_inode_shortcuts(e2fsck_t ctx, int bool)
 		fs->write_inode = 0;
 	}
 }
-
-		
