@@ -65,7 +65,7 @@
  */
 struct dup_block {
 	blk_t		block;		/* Block number */
-	ino_t		ino;		/* Inode number */
+	ext2_ino_t	ino;		/* Inode number */
 	int		num_bad;
 	/* Pointer to next dup record with different block */
 	struct dup_block *next_block;
@@ -81,17 +81,20 @@ struct dup_block {
  * of multiply-claimed blocks.
  */
 struct dup_inode {
-	ino_t			ino, dir;
+	ext2_ino_t		ino, dir;
 	int			num_dupblocks;
 	struct ext2_inode	inode;
 	struct dup_inode	*next;
 };
 
 static int process_pass1b_block(ext2_filsys fs, blk_t	*blocknr,
-				int	blockcnt, void	*priv_data);
+				e2_blkcnt_t blockcnt, blk_t ref_blk, 
+				int ref_offset, void *priv_data);
 static void delete_file(e2fsck_t ctx, struct dup_inode *dp,
 			char *block_buf);
 static int clone_file(e2fsck_t ctx, struct dup_inode *dp, char* block_buf);
+static int check_if_fs_block(e2fsck_t ctx, blk_t test_blk);
+
 static void pass1b(e2fsck_t ctx, char *block_buf);
 static void pass1c(e2fsck_t ctx, char *block_buf);
 static void pass1d(e2fsck_t ctx, char *block_buf);
@@ -115,7 +118,7 @@ void e2fsck_pass1_dupblocks(e2fsck_t ctx, char *block_buf)
 	clear_problem_context(&pctx);
 	
 	pctx.errcode = ext2fs_allocate_inode_bitmap(fs,
-		      "multiply claimed inode map", &inode_dup_map);
+		      _("multiply claimed inode map"), &inode_dup_map);
 	if (pctx.errcode) {
 		fix_problem(ctx, PR_1B_ALLOCATE_IBITMAP_ERROR, &pctx);
 		ctx->flags |= E2F_FLAG_ABORT;
@@ -149,19 +152,18 @@ void e2fsck_pass1_dupblocks(e2fsck_t ctx, char *block_buf)
  * Scan the inodes looking for inodes that contain duplicate blocks.
  */
 struct process_block_struct {
-	ino_t	ino;
-	int	dup_blocks;
-	e2fsck_t ctx;
+	ext2_ino_t	ino;
+	int		dup_blocks;
+	e2fsck_t	ctx;
 	struct problem_context *pctx;
 };
 
 static void pass1b(e2fsck_t ctx, char *block_buf)
 {
 	ext2_filsys fs = ctx->fs;
-	ino_t	ino;
+	ext2_ino_t ino;
 	struct ext2_inode inode;
 	ext2_inode_scan	scan;
-	errcode_t	retval;
 	struct process_block_struct pb;
 	struct dup_inode *dp;
 	struct problem_context pctx;
@@ -185,6 +187,7 @@ static void pass1b(e2fsck_t ctx, char *block_buf)
 	ctx->stashed_inode = &inode;
 	pb.ctx = ctx;
 	pb.pctx = &pctx;
+	pctx.str = "pass1b";
 	while (ino) {
 		pctx.ino = ctx->stashed_ino = ino;
 		if ((ino != EXT2_BAD_INO) &&
@@ -194,7 +197,7 @@ static void pass1b(e2fsck_t ctx, char *block_buf)
 
 		pb.ino = ino;
 		pb.dup_blocks = 0;
-		retval = ext2fs_block_iterate(fs, ino, 0, block_buf,
+		pctx.errcode = ext2fs_block_iterate2(fs, ino, 0, block_buf,
 					      process_pass1b_block, &pb);
 		if (pb.dup_blocks) {
 			end_problem_latch(ctx, PR_LATCH_DBLOCK);
@@ -210,10 +213,8 @@ static void pass1b(e2fsck_t ctx, char *block_buf)
 			if (ino != EXT2_BAD_INO)
 				dup_inode_count++;
 		}
-		if (retval)
-			com_err(ctx->program_name, retval,
-				"while calling ext2fs_block_iterate in pass1b");
-		
+		if (pctx.errcode)
+			fix_problem(ctx, PR_1B_BLOCK_ITERATE, &pctx);
 	next:
 		pctx.errcode = ext2fs_get_next_inode(scan, &ino, &inode);
 		if (pctx.errcode == EXT2_ET_BAD_BLOCK_IN_INODE_TABLE)
@@ -225,21 +226,22 @@ static void pass1b(e2fsck_t ctx, char *block_buf)
 		}
 	}
 	ext2fs_close_inode_scan(scan);
-	fs->get_blocks = 0;
-	fs->check_directory = 0;
+	e2fsck_use_inode_shortcuts(ctx, 0);
 }
 
-int process_pass1b_block(ext2_filsys fs,
-			 blk_t	*block_nr,
-			 int blockcnt,
-			 void *priv_data)
+static int process_pass1b_block(ext2_filsys fs,
+				blk_t	*block_nr,
+				e2_blkcnt_t blockcnt,
+				blk_t ref_blk, 
+				int ref_offset, 			 
+				void *priv_data)
 {
 	struct process_block_struct *p;
 	struct dup_block *dp, *q, *r;
 	int i;
 	e2fsck_t ctx;
 
-	if (!*block_nr)
+	if (HOLE_BLKADDR(*block_nr))
 		return 0;
 	p = (struct process_block_struct *) priv_data;
 	ctx = p->ctx;
@@ -292,11 +294,11 @@ int process_pass1b_block(ext2_filsys fs,
  */
 struct search_dir_struct {
 	int		count;
-	ino_t		first_inode;
-	ino_t		max_inode;
+	ext2_ino_t	first_inode;
+	ext2_ino_t	max_inode;
 };
 
-static int search_dirent_proc(ino_t dir, int entry,
+static int search_dirent_proc(ext2_ino_t dir, int entry,
 			      struct ext2_dir_entry *dirent,
 			      int offset, int blocksize,
 			      char *buf, void *priv_data)
@@ -369,7 +371,7 @@ static void pass1d(e2fsck_t ctx, char *block_buf)
 	ext2_filsys fs = ctx->fs;
 	struct dup_inode	*p, *s;
 	struct dup_block	*q, *r;
-	ino_t	*shared;
+	ext2_ino_t		*shared;
 	int	shared_len;
 	int	i;
 	int	file_ok;
@@ -383,8 +385,8 @@ static void pass1d(e2fsck_t ctx, char *block_buf)
 
 	pctx.num = dup_inode_count;
 	fix_problem(ctx, PR_1D_NUM_DUP_INODES, &pctx);
-	shared = (ino_t *) e2fsck_allocate_memory(ctx,
-				sizeof(ino_t) * dup_inode_count,
+	shared = (ext2_ino_t *) e2fsck_allocate_memory(ctx,
+				sizeof(ext2_ino_t) * dup_inode_count,
 				"Shared inode list");
 	for (p = dup_ino; p; p = p->next) {
 		shared_len = 0;
@@ -408,8 +410,7 @@ static void pass1d(e2fsck_t ctx, char *block_buf)
 				continue;
 			if (q->num_bad > 1)
 				file_ok = 0;
-			if (ext2fs_test_block_bitmap(ctx->block_illegal_map,
-						     q->block)) {
+			if (check_if_fs_block(ctx, q->block)) {
 				file_ok = 0;
 				meta_data = 1;
 			}
@@ -482,7 +483,9 @@ static void pass1d(e2fsck_t ctx, char *block_buf)
 
 static int delete_file_block(ext2_filsys fs,
 			     blk_t	*block_nr,
-			     int blockcnt,
+			     e2_blkcnt_t blockcnt,
+			     blk_t ref_block,
+			     int ref_offset, 
 			     void *priv_data)
 {
 	struct process_block_struct *pb;
@@ -492,7 +495,7 @@ static int delete_file_block(ext2_filsys fs,
 	pb = (struct process_block_struct *) priv_data;
 	ctx = pb->ctx;
 
-	if (!*block_nr)
+	if (HOLE_BLKADDR(*block_nr))
 		return 0;
 
 	if (ext2fs_test_block_bitmap(ctx->block_dup_map, *block_nr)) {
@@ -506,7 +509,7 @@ static int delete_file_block(ext2_filsys fs,
 							   *block_nr);
 		} else
 			com_err("delete_file_block", 0,
-				"internal error; can't find dup_blk for %d\n",
+			    _("internal error; can't find dup_blk for %d\n"),
 				*block_nr);
 	} else {
 		ext2fs_unmark_block_bitmap(ctx->block_found_map, *block_nr);
@@ -519,20 +522,20 @@ static int delete_file_block(ext2_filsys fs,
 static void delete_file(e2fsck_t ctx, struct dup_inode *dp, char* block_buf)
 {
 	ext2_filsys fs = ctx->fs;
-	errcode_t	retval;
 	struct process_block_struct pb;
 	struct ext2_inode	inode;
+	struct problem_context	pctx;
 
-	pb.ino = dp->ino;
+	clear_problem_context(&pctx);
+	pctx.ino = pb.ino = dp->ino;
 	pb.dup_blocks = dp->num_dupblocks;
 	pb.ctx = ctx;
-	
-	retval = ext2fs_block_iterate(fs, dp->ino, 0, block_buf,
-				      delete_file_block, &pb);
-	if (retval)
-		com_err("delete_file", retval,
-			"while calling ext2fs_block_iterate for inode %d",
-			dp->ino);
+	pctx.str = "delete_file";
+
+	pctx.errcode = ext2fs_block_iterate2(fs, dp->ino, 0, block_buf,
+				       delete_file_block, &pb);
+	if (pctx.errcode)
+		fix_problem(ctx, PR_1B_BLOCK_ITERATE, &pctx);
 	ext2fs_unmark_inode_bitmap(ctx->inode_used_map, dp->ino);
 	ext2fs_unmark_inode_bitmap(ctx->inode_dir_map, dp->ino);
 	if (ctx->inode_bad_map)
@@ -548,14 +551,16 @@ static void delete_file(e2fsck_t ctx, struct dup_inode *dp, char* block_buf)
 
 struct clone_struct {
 	errcode_t	errcode;
-	ino_t		dir;
+	ext2_ino_t	dir;
 	char	*buf;
 	e2fsck_t ctx;
 };
 
 static int clone_file_block(ext2_filsys fs,
 			    blk_t	*block_nr,
-			    int blockcnt,
+			    e2_blkcnt_t blockcnt,
+			    blk_t ref_block,
+			    int ref_offset, 
 			    void *priv_data)
 {
 	struct dup_block *p;
@@ -566,7 +571,7 @@ static int clone_file_block(ext2_filsys fs,
 
 	ctx = cs->ctx;
 	
-	if (!*block_nr)
+	if (HOLE_BLKADDR(*block_nr))
 		return 0;
 
 	if (ext2fs_test_block_bitmap(ctx->block_dup_map, *block_nr)) {
@@ -580,7 +585,7 @@ static int clone_file_block(ext2_filsys fs,
 				cs->errcode = retval;
 				return BLOCK_ABORT;
 			}
-			if (cs->dir) {
+			if (cs->dir && (blockcnt >= 0)) {
 				retval = ext2fs_set_dir_block(fs->dblist,
 				      cs->dir, new_block, blockcnt);
 				if (retval) {
@@ -602,8 +607,7 @@ static int clone_file_block(ext2_filsys fs,
 			}
 			p->num_bad--;
 			if (p->num_bad == 1 &&
-			    !ext2fs_test_block_bitmap(ctx->block_illegal_map,
-						      *block_nr))
+			    !check_if_fs_block(ctx, *block_nr))
 				ext2fs_unmark_block_bitmap(ctx->block_dup_map,
 							   *block_nr);
 			*block_nr = new_block;
@@ -613,7 +617,7 @@ static int clone_file_block(ext2_filsys fs,
 			return BLOCK_CHANGED;
 		} else
 			com_err("clone_file_block", 0,
-				"internal error; can't find dup_blk for %d\n",
+			    _("internal error; can't find dup_blk for %d\n"),
 				*block_nr);
 	}
 	return 0;
@@ -624,7 +628,9 @@ static int clone_file(e2fsck_t ctx, struct dup_inode *dp, char* block_buf)
 	ext2_filsys fs = ctx->fs;
 	errcode_t	retval;
 	struct clone_struct cs;
+	struct problem_context	pctx;
 
+	clear_problem_context(&pctx);
 	cs.errcode = 0;
 	cs.dir = 0;
 	cs.ctx = ctx;
@@ -634,21 +640,58 @@ static int clone_file(e2fsck_t ctx, struct dup_inode *dp, char* block_buf)
 
 	if (ext2fs_test_inode_bitmap(ctx->inode_dir_map, dp->ino))
 		cs.dir = dp->ino;
-	
-	retval = ext2fs_block_iterate(fs, dp->ino, 0, block_buf,
+
+	pctx.ino = dp->ino;
+	pctx.str = "clone_file";
+	pctx.errcode = ext2fs_block_iterate2(fs, dp->ino, 0, block_buf,
 				      clone_file_block, &cs);
 	ext2fs_mark_bb_dirty(fs);
 	ext2fs_free_mem((void **) &cs.buf);
-	if (retval) {
-		com_err("clone_file", retval,
-			"while calling ext2fs_block_iterate for inode %d",
-			dp->ino);
-		return retval;
+	if (pctx.errcode) {
+		fix_problem(ctx, PR_1B_BLOCK_ITERATE, &pctx);
+		return pctx.errcode;
 	}
 	if (cs.errcode) {
 		com_err("clone_file", cs.errcode,
-			"returned from clone_file_block");
-		return retval;
+			_("returned from clone_file_block"));
+		return cs.errcode;
+	}
+	return 0;
+}
+
+/*
+ * This routine returns 1 if a block overlaps with one of the superblocks,
+ * group descriptors, inode bitmaps, or block bitmaps.
+ */
+static int check_if_fs_block(e2fsck_t ctx, blk_t test_block)
+{
+	ext2_filsys fs = ctx->fs;
+	blk_t	block;
+	int	i;
+	
+	block = fs->super->s_first_data_block;
+	for (i = 0; i < fs->group_desc_count; i++) {
+
+		/* Check superblocks/block group descriptros */
+		if (ext2fs_bg_has_super(fs, i)) {
+			if (test_block >= block &&
+			    (test_block <= block + fs->desc_blocks))
+				return 1;
+		}
+		
+		/* Check the inode table */
+		if ((fs->group_desc[i].bg_inode_table) &&
+		    (test_block >= fs->group_desc[i].bg_inode_table) &&
+		    (test_block < (fs->group_desc[i].bg_inode_table +
+				   fs->inode_blocks_per_group)))
+			return 1;
+
+		/* Check the bitmap blocks */
+		if ((test_block == fs->group_desc[i].bg_block_bitmap) ||
+		    (test_block == fs->group_desc[i].bg_inode_bitmap))
+			return 1;
+		
+		block += fs->super->s_blocks_per_group;
 	}
 	return 0;
 }
