@@ -53,8 +53,7 @@
 /*
  * Keeps track of how many times an inode is referenced.
  */
-static void deallocate_inode(e2fsck_t ctx, ino_t ino,
-			     char* block_buf);
+static void deallocate_inode(e2fsck_t ctx, ext2_ino_t ino, char* block_buf);
 static int check_dir_block(ext2_filsys fs,
 			   struct ext2_db_entry *dir_blocks_info,
 			   void *priv_data);
@@ -63,8 +62,10 @@ static int allocate_dir_block(e2fsck_t ctx,
 			      char *buf, struct problem_context *pctx);
 static int update_dir_block(ext2_filsys fs,
 			    blk_t	*block_nr,
-			    int blockcnt,
-			    void *priv_data);
+			    e2_blkcnt_t blockcnt,
+			    blk_t	ref_block,
+			    int		ref_offset, 
+			    void	*priv_data);
 
 struct check_dir_struct {
 	char *buf;
@@ -159,7 +160,7 @@ void e2fsck_pass2(e2fsck_t ctx)
  */
 static int check_dot(e2fsck_t ctx,
 		     struct ext2_dir_entry *dirent,
-		     ino_t ino, struct problem_context *pctx)
+		     ext2_ino_t ino, struct problem_context *pctx)
 {
 	struct ext2_dir_entry *nextdir;
 	int	status = 0;
@@ -259,7 +260,7 @@ static int check_dotdot(e2fsck_t ctx,
  */
 static int check_name(e2fsck_t ctx,
 		      struct ext2_dir_entry *dirent,
-		      ino_t dir_ino, struct problem_context *pctx)
+		      ext2_ino_t dir_ino, struct problem_context *pctx)
 {
 	int	i;
 	int	fixup = -1;
@@ -284,7 +285,7 @@ static int check_name(e2fsck_t ctx,
  */
 static _INLINE_ int check_filetype(e2fsck_t ctx,
 		      struct ext2_dir_entry *dirent,
-		      ino_t dir_ino, struct problem_context *pctx)
+		      ext2_ino_t dir_ino, struct problem_context *pctx)
 {
 	int	filetype = dirent->name_len >> 8;
 	int	should_be = EXT2_FT_UNKNOWN;
@@ -336,7 +337,7 @@ static int check_dir_block(ext2_filsys fs,
 	int			dir_modified = 0;
 	int			dot_state;
 	blk_t			block_nr = db->blk;
-	ino_t 			ino = db->ino;
+	ext2_ino_t 		ino = db->ino;
 	__u16			links;
 	struct check_dir_struct	*cd;
 	char 			*buf;
@@ -382,6 +383,8 @@ static int check_dir_block(ext2_filsys fs,
 #endif
 	
 	cd->pctx.errcode = ext2fs_read_dir_block(fs, block_nr, buf);
+	if (cd->pctx.errcode == EXT2_ET_DIR_CORRUPTED)
+		cd->pctx.errcode = 0; /* We'll handle this ourselves */
 	if (cd->pctx.errcode) {
 		if (!fix_problem(ctx, PR_2_READ_DIRBLOCK, &cd->pctx)) {
 			ctx->flags |= E2F_FLAG_ABORT;
@@ -599,13 +602,15 @@ static int check_dir_block(ext2_filsys fs,
  * functioned called by deallocate inode via ext2fs_iterate_block().
  */
 static int deallocate_inode_block(ext2_filsys fs,
-			     blk_t	*block_nr,
-			     int blockcnt,
-			     void *priv_data)
+				  blk_t	*block_nr,
+				  e2_blkcnt_t blockcnt,
+				  blk_t ref_block,
+				  int ref_offset, 
+				  void *priv_data)
 {
 	e2fsck_t	ctx = (e2fsck_t) priv_data;
 	
-	if (!*block_nr)
+	if (HOLE_BLKADDR(*block_nr))
 		return 0;
 	ext2fs_unmark_block_bitmap(ctx->block_found_map, *block_nr);
 	ext2fs_unmark_block_bitmap(fs->block_map, *block_nr);
@@ -615,8 +620,7 @@ static int deallocate_inode_block(ext2_filsys fs,
 /*
  * This fuction deallocates an inode
  */
-static void deallocate_inode(e2fsck_t ctx, ino_t ino,
-			     char* block_buf)
+static void deallocate_inode(e2fsck_t ctx, ext2_ino_t ino, char* block_buf)
 {
 	ext2_filsys fs = ctx->fs;
 	struct ext2_inode	inode;
@@ -645,7 +649,7 @@ static void deallocate_inode(e2fsck_t ctx, ino_t ino,
 		return;
 	
 	ext2fs_mark_bb_dirty(fs);
-	pctx.errcode = ext2fs_block_iterate(fs, ino, 0, block_buf,
+	pctx.errcode = ext2fs_block_iterate2(fs, ino, 0, block_buf,
 					    deallocate_inode_block, ctx);
 	if (pctx.errcode) {
 		fix_problem(ctx, PR_2_DEALLOC_INODE, &pctx);
@@ -654,7 +658,8 @@ static void deallocate_inode(e2fsck_t ctx, ino_t ino,
 	}
 }
 
-extern int e2fsck_process_bad_inode(e2fsck_t ctx, ino_t dir, ino_t ino)
+extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
+				    ext2_ino_t ino)
 {
 	ext2_filsys fs = ctx->fs;
 	struct ext2_inode	inode;
@@ -675,22 +680,22 @@ extern int e2fsck_process_bad_inode(e2fsck_t ctx, ino_t dir, ino_t ino)
 	    !LINUX_S_ISLNK(inode.i_mode) && !LINUX_S_ISFIFO(inode.i_mode) &&
 	    !(LINUX_S_ISSOCK(inode.i_mode)))
 		problem = PR_2_BAD_MODE;
-
-	if (LINUX_S_ISCHR(inode.i_mode)
-	    && !e2fsck_pass1_check_device_inode(&inode))
+	else if (LINUX_S_ISCHR(inode.i_mode)
+		 && !e2fsck_pass1_check_device_inode(&inode))
 		problem = PR_2_BAD_CHAR_DEV;
-		
-	if (LINUX_S_ISBLK(inode.i_mode)
-	    && !e2fsck_pass1_check_device_inode(&inode))
+	else if (LINUX_S_ISBLK(inode.i_mode)
+		 && !e2fsck_pass1_check_device_inode(&inode))
 		problem = PR_2_BAD_BLOCK_DEV;
-
-	if (LINUX_S_ISFIFO(inode.i_mode)
-	    && !e2fsck_pass1_check_device_inode(&inode))
+	else if (LINUX_S_ISFIFO(inode.i_mode)
+		 && !e2fsck_pass1_check_device_inode(&inode))
 		problem = PR_2_BAD_FIFO;
-		
-	if (LINUX_S_ISSOCK(inode.i_mode)
-	    && !e2fsck_pass1_check_device_inode(&inode))
+	else if (LINUX_S_ISSOCK(inode.i_mode)
+		 && !e2fsck_pass1_check_device_inode(&inode))
 		problem = PR_2_BAD_SOCKET;
+	else if (LINUX_S_ISLNK(inode.i_mode)
+		 && (inode.i_size_high ||
+		     (inode.i_size > EXT2_N_BLOCKS*4)))
+		problem = PR_2_SYMLINK_SIZE;
 
 	if (problem) {
 		if (fix_problem(ctx, problem, &pctx)) {
@@ -831,7 +836,7 @@ static int allocate_dir_block(e2fsck_t ctx,
 	 * Finally, update the block pointers for the inode
 	 */
 	db->blk = blk;
-	pctx->errcode = ext2fs_block_iterate(fs, db->ino, BLOCK_FLAG_HOLE,
+	pctx->errcode = ext2fs_block_iterate2(fs, db->ino, BLOCK_FLAG_HOLE,
 				      0, update_dir_block, db);
 	if (pctx->errcode) {
 		pctx->str = "ext2fs_block_iterate";
@@ -847,13 +852,15 @@ static int allocate_dir_block(e2fsck_t ctx,
  */
 static int update_dir_block(ext2_filsys fs,
 			    blk_t	*block_nr,
-			    int blockcnt,
+			    e2_blkcnt_t blockcnt,
+			    blk_t ref_block,
+			    int ref_offset, 
 			    void *priv_data)
 {
 	struct ext2_db_entry *db;
 
 	db = (struct ext2_db_entry *) priv_data;
-	if (db->blockcnt == blockcnt) {
+	if (db->blockcnt == (int) blockcnt) {
 		*block_nr = db->blk;
 		return BLOCK_CHANGED;
 	}
