@@ -22,6 +22,9 @@
 #endif
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#else
+extern char *optarg;
+extern int optind;
 #endif
 #include <unistd.h>
 #ifdef HAVE_ERRNO_H
@@ -55,26 +58,29 @@ static int possible_block_sizes[] = { 1024, 2048, 4096, 8192, 0};
 static int root_filesystem = 0;
 static int read_only_root = 0;
 
+e2fsck_t e2fsck_global_ctx;	/* Try your very best not to use this! */
+
 static void usage(e2fsck_t ctx)
 {
 	fprintf(stderr,
-		"Usage: %s [-panyrcdfvstFSV] [-b superblock] [-B blocksize]\n"
+		_("Usage: %s [-panyrcdfvstFSV] [-b superblock] [-B blocksize]\n"
 		"\t\t[-I inode_buffer_blocks] [-P process_inode_size]\n"
-		"\t\t[-l|-L bad_blocks_file] [-C fd] device\n",
+		"\t\t[-l|-L bad_blocks_file] [-C fd] device\n"),
 		ctx->program_name);
 
-	fprintf(stderr, "\nEmergency help:\n"
+	fprintf(stderr, _("\nEmergency help:\n"
 		" -p                   Automatic repair (no questions)\n"
 		" -n                   Make no changes to the filesystem\n"
 		" -y                   Assume \"yes\" to all questions\n"
 		" -c                   Check for bad blocks\n"
-		" -f                   Force checking even if filesystem is marked clean\n"
+		" -f                   Force checking even if filesystem is marked clean\n"));
+	fprintf(stderr, _(""
 		" -v                   Be verbose\n"
 		" -b superblock        Use alternative superblock\n"
 		" -B blocksize         Force blocksize when looking for superblock\n"
 		" -l bad_blocks_file   Add to badblocks list\n"
 		" -L bad_blocks_file   Set badblocks list\n"
-		);
+		));
 
 	exit(FSCK_USAGE);
 }
@@ -101,12 +107,51 @@ static void show_stats(e2fsck_t	ctx)
 	frag_percent = (frag_percent + 5) / 10;
 	
 	if (!verbose) {
-		printf("%s: %d/%d files (%0d.%d%% non-contiguous), %d/%d blocks\n",
+		printf(_("%s: %d/%d files (%0d.%d%% non-contiguous), %d/%d blocks\n"),
 		       ctx->device_name, inodes_used, inodes,
 		       frag_percent / 10, frag_percent % 10,
 		       blocks_used, blocks);
 		return;
 	}
+	/*
+	 * This is a bit ugly. But I think there will nearly always be more
+	 * than one "thing" to report about, so I won't try writing complex
+	 * code to handle one/two/many forms of all words.
+	 * Some languages (Italian, at least) never uses the plural form
+	 * of foreign words, so in real life this could not be a problem.
+	 * md@linux.it - 2000-1-15
+	 */
+#ifdef ENABLE_NLS
+	printf (_("\n%8d inodes used (%d%%)\n"), inodes_used,
+		(inodes_used != 1), 100 * inodes_used / inodes);
+	printf (_("%8d non-contiguous inodes (%0d.%d%%)\n"),
+		ctx->fs_fragmented, frag_percent / 10, frag_percent % 10);
+	printf (_("         # of inodes with ind/dind/tind blocks: %d/%d/%d\n"),
+		ctx->fs_ind_count, ctx->fs_dind_count, ctx->fs_tind_count);
+	printf (_("%8d blocks used (%d%%)\n"
+		"%8d bad blocks\n"), blocks_used,
+		100 * blocks_used / blocks, ctx->fs_badblocks_count);
+	printf (_("\n%8d regular files\n"
+		"%8d directories\n"
+		"%8d character device files\n"
+		"%8d block device files\n"
+		"%8d fifos\n"
+		"%8d links\n"
+		"%8d symbolic links (%d fast symbolic links)\n"
+		"%8d sockets\n"
+		"--------\n"
+		"%8d files\n"),
+		ctx->fs_regular_count,
+		ctx->fs_directory_count,
+		ctx->fs_chardev_count,
+		ctx->fs_blockdev_count,
+		ctx->fs_fifo_count,
+		ctx->fs_links_count - dir_links,
+		ctx->fs_symlinks_count,
+		ctx->fs_fast_symlinks_count,
+		ctx->fs_sockets_count,
+		ctx->fs_total_count - dir_links);
+#else
 	printf ("\n%8d inode%s used (%d%%)\n", inodes_used,
 		(inodes_used != 1) ? "s" : "",
 		100 * inodes_used / inodes);
@@ -148,6 +193,7 @@ static void show_stats(e2fsck_t	ctx)
 		ctx->fs_sockets_count, (ctx->fs_sockets_count != 1) ? "s" : "",
 		ctx->fs_total_count - dir_links,
 		((ctx->fs_total_count - dir_links) != 1) ? "s" : "");
+#endif
 }
 
 static void check_mount(e2fsck_t ctx)
@@ -158,7 +204,7 @@ static void check_mount(e2fsck_t ctx)
 	retval = ext2fs_check_if_mounted(ctx->filesystem_name, &mount_flags);
 	if (retval) {
 		com_err("ext2fs_check_if_mount", retval,
-			"while determining whether %s is mounted.",
+			_("while determining whether %s is mounted."),
 			ctx->filesystem_name);
 		return;
 	}
@@ -180,21 +226,19 @@ static void check_mount(e2fsck_t ctx)
 #endif
 	
 	if (ctx->options & E2F_OPT_READONLY) {
-		printf("Warning!  %s is mounted.\n", ctx->filesystem_name);
+		printf(_("Warning!  %s is mounted.\n"), ctx->filesystem_name);
 		return;
 	}
 
-	printf("%s is mounted.  ", ctx->filesystem_name);
-	if (!isatty(0) || !isatty(1)) {
-		printf("Cannot continue, aborting.\n\n");
-		exit(FSCK_ERROR);
-	}
-	printf("\n\n\007\007\007\007WARNING!!!  "
+	printf(_("%s is mounted.  "), ctx->filesystem_name);
+	if (!isatty(0) || !isatty(1))
+		fatal_error(ctx, _("Cannot continue, aborting.\n\n"));
+	printf(_("\n\n\007\007\007\007WARNING!!!  "
 	       "Running e2fsck on a mounted filesystem may cause\n"
-	       "SEVERE filesystem damage.\007\007\007\n\n");
-	cont = ask_yn("Do you really want to continue", -1);
+	       "SEVERE filesystem damage.\007\007\007\n\n"));
+	cont = ask_yn(_("Do you really want to continue"), -1);
 	if (!cont) {
-		printf ("check aborted.\n");
+		printf (_("check aborted.\n"));
 		exit (0);
 	}
 	return;
@@ -209,26 +253,33 @@ static void check_if_skip(e2fsck_t ctx)
 {
 	ext2_filsys fs = ctx->fs;
 	const char *reason = NULL;
+	unsigned int reason_arg = 0;
 	
 	if (force || bad_blocks_file || cflag || swapfs)
 		return;
 	
 	if (fs->super->s_state & EXT2_ERROR_FS)
-		reason = "contains a file system with errors";
+		reason = _(" contains a file system with errors");
 	else if ((fs->super->s_state & EXT2_VALID_FS) == 0)
-		reason = "was not cleanly unmounted";
-	else if (fs->super->s_mnt_count >=
-		 (unsigned) fs->super->s_max_mnt_count)
-		reason = "has reached maximal mount count";
-	else if (fs->super->s_checkinterval &&
+		reason = _(" was not cleanly unmounted");
+	else if ((fs->super->s_max_mnt_count > 0) &&
+		 (fs->super->s_mnt_count >=
+		  (unsigned) fs->super->s_max_mnt_count)) {
+		reason = _(" has been mounted %u times without being checked");
+		reason_arg = fs->super->s_mnt_count;
+	} else if (fs->super->s_checkinterval &&
 		 time(0) >= (fs->super->s_lastcheck +
-			     fs->super->s_checkinterval))
-		reason = "has gone too long without being checked";
+			     fs->super->s_checkinterval)) {
+		reason = _(" has gone %u days without being checked");
+		reason_arg = (time(0) - fs->super->s_lastcheck)/(3600*24);
+	}
 	if (reason) {
-		printf("%s %s, check forced.\n", ctx->device_name, reason);
+		fputs(ctx->device_name, stdout);
+		printf(reason, reason_arg);
+		fputs(_(", check forced.\n"), stdout);
 		return;
 	}
-	printf("%s: clean, %d/%d files, %d/%d blocks\n", ctx->device_name,
+	printf(_("%s: clean, %d/%d files, %d/%d blocks\n"), ctx->device_name,
 	       fs->super->s_inodes_count - fs->super->s_free_inodes_count,
 	       fs->super->s_inodes_count,
 	       fs->super->s_blocks_count - fs->super->s_free_blocks_count,
@@ -247,7 +298,6 @@ struct percent_tbl {
 struct percent_tbl e2fsck_tbl = {
 	5, { 0, 70, 90, 92,  95, 100 }
 };
-static int dpywidth = 50;
 static char bar[] =
 	"==============================================================="
 	"===============================================================";
@@ -262,7 +312,7 @@ static float calc_percent(struct percent_tbl *tbl, int pass, int curr,
 	
 	if (pass <= 0)
 		return 0.0;
-	if (pass > tbl->max_pass)
+	if (pass > tbl->max_pass || max == 0)
 		return 100.0;
 	percent = ((float) curr) / ((float) max);
 	return ((percent * (tbl->table[pass] - tbl->table[pass-1]))
@@ -281,10 +331,13 @@ extern void e2fsck_clear_progbar(e2fsck_t ctx)
 static int e2fsck_update_progress(e2fsck_t ctx, int pass,
 				  unsigned long cur, unsigned long max)
 {
-	const char spinner[] = "\\|/-";
+	static const char spinner[] = "\\|/-";
 	char buf[80];
 	int	i;
 	float percent;
+	int	tick;
+	struct timeval	tv;
+	static int dpywidth = 0;
 
 	if (pass == 0)
 		return 0;
@@ -295,12 +348,39 @@ static int e2fsck_update_progress(e2fsck_t ctx, int pass,
 	} else {
 		if (ctx->flags & E2F_FLAG_PROG_SUPPRESS)
 			return 0;
+		if (dpywidth == 0) {
+			dpywidth = 66 - strlen(ctx->device_name);
+			dpywidth = 8 * (dpywidth / 8);
+		}
+		/*
+		 * Calculate the new progress position.  If the
+		 * percentage hasn't changed, then we skip out right
+		 * away. 
+		 */
+		percent = calc_percent(&e2fsck_tbl, pass, cur, max);
+		if (ctx->progress_last_percent == (int) 10 * percent)
+			return 0;
+		ctx->progress_last_percent = (int) 10 * percent;
+
+		/*
+		 * If we've already updated the spinner once within
+		 * the last 1/8th of a second, no point doing it
+		 * again.
+		 */
+		gettimeofday(&tv, NULL);
+		tick = (tv.tv_sec << 3) + (tv.tv_usec / (1000000 / 8));
+		if ((tick == ctx->progress_last_time) &&
+		    (cur != max) && (cur != 0))
+			return 0;
+		ctx->progress_last_time = tick;
+
+		/*
+		 * Advance the spinner, and note that the progress bar
+		 * will be on the screen
+		 */
 		ctx->progress_pos = (ctx->progress_pos+1) & 3;
 		ctx->flags |= E2F_FLAG_PROG_BAR;
-		percent = calc_percent(&e2fsck_tbl, pass, cur, max);
-		if (ctx->progress_last_percent == (int) 1000 * percent)
-			return 0;
-		ctx->progress_last_percent = (int) 1000 * percent;
+		
 		i = ((percent * dpywidth) + 50) / 100;
 		printf("%s: |%s%s", ctx->device_name,
 		       bar + (sizeof(bar) - (i+1)),
@@ -319,7 +399,7 @@ static int e2fsck_update_progress(e2fsck_t ctx, int pass,
 
 #define PATH_SET "PATH=/sbin"
 
-static void reserve_stdio_fds(NOARGS)
+static void reserve_stdio_fds(void)
 {
 	int	fd;
 
@@ -328,8 +408,8 @@ static void reserve_stdio_fds(NOARGS)
 		if (fd > 2)
 			break;
 		if (fd < 0) {
-			fprintf(stderr, "ERROR: Couldn't open "
-				"/dev/null (%s)\n",
+			fprintf(stderr, _("ERROR: Couldn't open "
+				"/dev/null (%s)\n"),
 				strerror(errno));
 			break;
 		}
@@ -338,11 +418,9 @@ static void reserve_stdio_fds(NOARGS)
 }
 
 #ifdef HAVE_SIGNAL_H
-static e2fsck_t global_signal_ctx;
-
 static void signal_progress_on(int sig)
 {
-	e2fsck_t ctx = global_signal_ctx;
+	e2fsck_t ctx = e2fsck_global_ctx;
 
 	if (!ctx)
 		return;
@@ -353,7 +431,7 @@ static void signal_progress_on(int sig)
 
 static void signal_progress_off(int sig)
 {
-	e2fsck_t ctx = global_signal_ctx;
+	e2fsck_t ctx = e2fsck_global_ctx;
 
 	if (!ctx)
 		return;
@@ -366,7 +444,7 @@ static void signal_progress_off(int sig)
 static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 {
 	int		flush = 0;
-	int		c;
+	int		c, fd;
 #ifdef MTRACE
 	extern void	*mallwatch;
 #endif
@@ -411,6 +489,19 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 		case 'C':
 			ctx->progress = e2fsck_update_progress;
 			ctx->progress_fd = atoi(optarg);
+			if (!ctx->progress_fd)
+				break;
+			/* Validate the file descriptor to avoid disasters */
+			fd = dup(ctx->progress_fd);
+			if (fd < 0) {
+				fprintf(stderr,
+				_("Error validating file descriptor %d: %s\n"),
+					ctx->progress_fd,
+					error_message(errno));
+				fatal_error(ctx,
+			_("Invalid completion information file descriptor"));
+			} else
+				close(fd);
 			break;
 		case 'p':
 		case 'a':
@@ -432,8 +523,8 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 			else
 				ctx->options |= E2F_OPT_TIME;
 #else
-			fprintf(stderr, "The -t option is not "
-				"supported on this version of e2fsck.\n");
+			fprintf(stderr, _("The -t option is not "
+				"supported on this version of e2fsck.\n"));
 #endif
 			break;
 		case 'c':
@@ -471,11 +562,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 			force = 1;
 			break;
 		case 'F':
-#ifdef BLKFLSBUF
 			flush = 1;
-#else
-			fatal_error(ctx, "-F not supported");
-#endif
 			break;
 		case 'v':
 			verbose = 1;
@@ -491,11 +578,20 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 		case 'N':
 			ctx->device_name = optarg;
 			break;
+#ifdef ENABLE_SWAPFS
 		case 's':
 			normalize_swapfs = 1;
 		case 'S':
 			swapfs = 1;
 			break;
+#else
+		case 's':
+		case 'S':
+			fprintf(stderr, _("Byte-swapping filesystems "
+					  "not compiled in this version "
+					  "of e2fsck\n"));
+			exit(1);
+#endif
 		default:
 			usage(ctx);
 		}
@@ -508,31 +604,31 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 		ctx->options |= E2F_OPT_READONLY;
 	ctx->filesystem_name = argv[optind];
 	if (flush) {
-#ifdef BLKFLSBUF
 		int	fd = open(ctx->filesystem_name, O_RDONLY, 0);
 
 		if (fd < 0) {
-			com_err("open", errno, "while opening %s for flushing",
+			com_err("open", errno,
+				_("while opening %s for flushing"),
 				ctx->filesystem_name);
-			exit(FSCK_ERROR);
+			fatal_error(ctx, 0);
 		}
-		if (ioctl(fd, BLKFLSBUF, 0) < 0) {
-			com_err("BLKFLSBUF", errno, "while trying to flush %s",
+		if ((retval = ext2fs_sync_device(fd, 1))) {
+			com_err("ext2fs_sync_device", retval,
+				_("while trying to flush %s"),
 				ctx->filesystem_name);
-			exit(FSCK_ERROR);
+			fatal_error(ctx, 0);
 		}
 		close(fd);
-#else
-		fatal_error(ctx, "BLKFLSBUF not supported");
-#endif /* BLKFLSBUF */
 	}
+#ifdef ENABLE_SWAPFS
 	if (swapfs) {
 		if (cflag || bad_blocks_file) {
-			fprintf(stderr, "Incompatible options not "
-				"allowed when byte-swapping.\n");
-			exit(FSCK_ERROR);
+			fprintf(stderr, _("Incompatible options not "
+				"allowed when byte-swapping.\n"));
+			exit(FSCK_USAGE);
 		}
 	}
+#endif
 #ifdef HAVE_SIGNAL_H
 	/*
 	 * Set up signal action
@@ -541,7 +637,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 #ifdef SA_RESTART
 	sa.sa_flags = SA_RESTART;
 #endif
-	global_signal_ctx = ctx;
+	e2fsck_global_ctx = ctx;
 	sa.sa_handler = signal_progress_on;
 	sigaction(SIGUSR1, &sa, 0);
 	sa.sa_handler = signal_progress_off;
@@ -560,7 +656,7 @@ int main (int argc, char *argv[])
 	int		i;
 	ext2_filsys	fs = 0;
 	io_manager	io_ptr;
-	struct ext2fs_sb *s;
+	struct ext2_super_block *sb;
 	const char	*lib_ver_date;
 	int		my_ver, lib_ver;
 	e2fsck_t	ctx;
@@ -573,20 +669,25 @@ int main (int argc, char *argv[])
 #endif
 #ifdef MCHECK
 	mcheck(0);
+#endif
+#ifdef ENABLE_NLS
+	setlocale(LC_MESSAGES, "");
+	bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
+	textdomain(NLS_CAT_NAME);
 #endif
 	my_ver = ext2fs_parse_version_string(my_ver_string);
 	lib_ver = ext2fs_get_library_version(0, &lib_ver_date);
 	if (my_ver > lib_ver) {
-		fprintf( stderr, "Error: ext2fs library version "
-			"out of date!\n");
+		fprintf( stderr, _("Error: ext2fs library version "
+			"out of date!\n"));
 		show_version_only++;
 	}
 	
 	retval = PRS(argc, argv, &ctx);
 	if (retval) {
 		com_err("e2fsck", retval,
-			"while trying to initialize program");
-		exit(1);
+			_("while trying to initialize program"));
+		exit(FSCK_ERROR);
 	}
 	reserve_stdio_fds();
 	
@@ -595,14 +696,14 @@ int main (int argc, char *argv[])
 #endif
 
 	if (!(ctx->options & E2F_OPT_PREEN) || show_version_only)
-		fprintf (stderr, "e2fsck %s, %s for EXT2 FS %s, %s\n",
+		fprintf (stderr, _("e2fsck %s, %s for EXT2 FS %s, %s\n"),
 			 my_ver_string, my_ver_date, EXT2FS_VERSION,
 			 EXT2FS_DATE);
 
 	if (show_version_only) {
-		fprintf(stderr, "\tUsing %s, %s\n",
+		fprintf(stderr, _("\tUsing %s, %s\n"),
 			error_message(EXT2_ET_BASE), lib_ver_date);
-		exit(0);
+		exit(FSCK_OK);
 	}
 	
 	check_mount(ctx);
@@ -612,7 +713,7 @@ int main (int argc, char *argv[])
 	    !(ctx->options & E2F_OPT_YES)) {
 		if (!isatty (0) || !isatty (1))
 			fatal_error(ctx,
-				    "need terminal for interactive repairs");
+				    _("need terminal for interactive repairs"));
 	}
 	ctx->superblock = ctx->use_superblock;
 restart:
@@ -622,7 +723,10 @@ restart:
 	io_ptr = test_io_manager;
 	test_io_backing_manager = unix_io_manager;
 #endif
-	flags = (ctx->options & E2F_OPT_READONLY) ? 0 : EXT2_FLAG_RW;
+	flags = 0;
+	if ((ctx->options & E2F_OPT_READONLY) == 0)
+		flags |= EXT2_FLAG_RW;
+
 	if (ctx->superblock && blocksize) {
 		retval = ext2fs_open(ctx->filesystem_name, flags,
 				     ctx->superblock, blocksize, io_ptr, &fs);
@@ -642,9 +746,9 @@ restart:
 	    ((retval == EXT2_ET_BAD_MAGIC) ||
 	     ((retval == 0) && ext2fs_check_desc(fs)))) {
 		if (!fs || (fs->group_desc_count > 1)) {
-			printf("%s trying backup blocks...\n",
-			       retval ? "Couldn't find ext2 superblock," :
-			       "Group descriptors look bad...");
+			printf(_("%s trying backup blocks...\n"),
+			       retval ? _("Couldn't find ext2 superblock,") :
+			       _("Group descriptors look bad..."));
 			ctx->superblock = get_backup_sb(fs);
 			if (fs)
 				ext2fs_close(fs);
@@ -652,72 +756,117 @@ restart:
 		}
 	}
 	if (retval) {
-		com_err(ctx->program_name, retval, "while trying to open %s",
+		com_err(ctx->program_name, retval, _("while trying to open %s"),
 			ctx->filesystem_name);
 		if (retval == EXT2_ET_REV_TOO_HIGH) {
-			printf("The filesystem revision is apparently "
+			printf(_("The filesystem revision is apparently "
 			       "too high for this version of e2fsck.\n"
 			       "(Or the filesystem superblock "
-			       "is corrupt)\n\n");
+			       "is corrupt)\n\n"));
 			fix_problem(ctx, PR_0_SB_CORRUPT, &pctx);
 		} else if (retval == EXT2_ET_SHORT_READ)
-			printf("Could this be a zero-length partition?\n");
+			printf(_("Could this be a zero-length partition?\n"));
 		else if ((retval == EPERM) || (retval == EACCES))
-			printf("You must have %s access to the "
-			       "filesystem or be root\n",
+			printf(_("You must have %s access to the "
+			       "filesystem or be root\n"),
 			       (ctx->options & E2F_OPT_READONLY) ?
 			       "r/o" : "r/w");
 		else if (retval == ENXIO)
-			printf("Possibly non-existent or swap device?\n");
+			printf(_("Possibly non-existent or swap device?\n"));
 #ifdef EROFS
 		else if (retval == EROFS)
-			printf("Disk write-protected; use the -n option "
+			printf(_("Disk write-protected; use the -n option "
 			       "to do a read-only\n"
-			       "check of the device.\n");
+			       "check of the device.\n"));
 #endif
 		else
 			fix_problem(ctx, PR_0_SB_CORRUPT, &pctx);
-		exit(FSCK_ERROR);
+		fatal_error(ctx, 0);
 	}
 	ctx->fs = fs;
 	fs->priv_data = ctx;
-#ifdef	EXT2_CURRENT_REV
-	if (fs->super->s_rev_level > E2FSCK_CURRENT_REV) {
+	sb = fs->super;
+	if (sb->s_rev_level > E2FSCK_CURRENT_REV) {
 		com_err(ctx->program_name, EXT2_ET_REV_TOO_HIGH,
-			"while trying to open %s",
+			_("while trying to open %s"),
 			ctx->filesystem_name);
 	get_newer:
-		fatal_error(ctx, "Get a newer version of e2fsck!");
+		fatal_error(ctx, _("Get a newer version of e2fsck!"));
 	}
-#endif
+
 	/*
-	 * Check for compatibility with the feature sets.  We need to
-	 * be more stringent than ext2fs_open().
+	 * Set the device name, which is used whenever we print error
+	 * or informational messages to the user.
 	 */
-	s = (struct ext2fs_sb *) fs->super;
-	if ((s->s_feature_compat & ~EXT2_LIB_FEATURE_COMPAT_SUPP) ||
-	    (s->s_feature_incompat & ~EXT2_LIB_FEATURE_INCOMPAT_SUPP)) {
-		com_err(ctx->program_name, EXT2_ET_UNSUPP_FEATURE,
-			"(%s)", ctx->filesystem_name);
-		goto get_newer;
-	}
-	if (s->s_feature_ro_compat & ~EXT2_LIB_FEATURE_RO_COMPAT_SUPP) {
-		com_err(ctx->program_name, EXT2_ET_RO_UNSUPP_FEATURE,
-			"(%s)", ctx->filesystem_name);
-		goto get_newer;
-	}
 	if (ctx->device_name == 0 &&
-	    (s->s_volume_name[0] != 0)) {
-		char *cp = malloc(sizeof(s->s_volume_name)+1);
+	    (sb->s_volume_name[0] != 0)) {
+		char *cp = malloc(sizeof(sb->s_volume_name)+1);
 		if (cp) {
-			strncpy(cp, s->s_volume_name,
-				sizeof(s->s_volume_name));
-			cp[sizeof(s->s_volume_name)] = 0;
+			strncpy(cp, sb->s_volume_name,
+				sizeof(sb->s_volume_name));
+			cp[sizeof(sb->s_volume_name)] = 0;
 			ctx->device_name = cp;
 		}
 	}
 	if (ctx->device_name == 0)
 		ctx->device_name = ctx->filesystem_name;
+
+	/*
+	 * Make sure the ext3 superblock fields are consistent.
+	 */
+	retval = e2fsck_check_ext3_journal(ctx);
+	if (retval) {
+		com_err(ctx->program_name, retval,
+			_("while checking ext3 journal for %s"),
+			ctx->device_name);
+		ext2fs_close(ctx->fs);
+		fatal_error(ctx, 0);
+	}
+
+	/*
+	 * Check to see if we need to do ext3-style recovery.  If so,
+	 * do it, and then restart the fsck.
+	 */
+	if (sb->s_feature_incompat & EXT3_FEATURE_INCOMPAT_RECOVER) {
+		if (ctx->options & E2F_OPT_READONLY) {
+			printf(_("Warning: skipping journal recovery "
+				 "because doing a read-only filesystem "
+				 "check.\n"));
+			io_channel_flush(ctx->fs->io);
+		} else {
+			retval = e2fsck_run_ext3_journal(ctx);
+			if (retval) {
+				com_err(ctx->program_name, retval,
+				_("while recovering ext3 journal of %s"),
+					ctx->device_name);
+				fatal_error(ctx, 0);
+			}
+			ext2fs_close(ctx->fs);
+			ctx->fs = 0;
+			goto restart;
+		}
+	}
+
+	/*
+	 * Check for compatibility with the feature sets.  We need to
+	 * be more stringent than ext2fs_open().
+	 */
+	if ((sb->s_feature_compat & ~EXT2_LIB_FEATURE_COMPAT_SUPP) ||
+	    (sb->s_feature_incompat & ~EXT2_LIB_FEATURE_INCOMPAT_SUPP)) {
+		com_err(ctx->program_name, EXT2_ET_UNSUPP_FEATURE,
+			"(%s)", ctx->device_name);
+		goto get_newer;
+	}
+	if (sb->s_feature_ro_compat & ~EXT2_LIB_FEATURE_RO_COMPAT_SUPP) {
+		com_err(ctx->program_name, EXT2_ET_RO_UNSUPP_FEATURE,
+			"(%s)", ctx->device_name);
+		goto get_newer;
+	}
+#ifdef ENABLE_COMPRESSION
+	if (sb->s_feature_incompat & EXT2_FEATURE_INCOMPAT_COMPRESSION)
+		com_err(ctx->program_name, 0,
+			_("Warning: compression support is experimental.\n"));
+#endif
 	
 	/*
 	 * If the user specified a specific superblock, presumably the
@@ -740,28 +889,29 @@ restart:
 		set_latch_flags(PR_LATCH_RELOC, PRL_LATCHED, 0);
 	check_super_block(ctx);
 	if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
-		exit(FSCK_ERROR);
+		fatal_error(ctx, 0);
 	check_if_skip(ctx);
 	if (bad_blocks_file)
 		read_bad_blocks_file(ctx, bad_blocks_file, replace_bad_blocks);
 	else if (cflag)
 		test_disk(ctx);
 	if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
-		exit(FSCK_ERROR);
-
+		fatal_error(ctx, 0);
+#ifdef ENABLE_SWAPFS
 	if (normalize_swapfs) {
 		if ((fs->flags & EXT2_FLAG_SWAP_BYTES) ==
 		    ext2fs_native_flag()) {
-			fprintf(stderr, "%s: Filesystem byte order "
-				"already normalized.\n", ctx->device_name);
-			exit(FSCK_ERROR);
+			fprintf(stderr, _("%s: Filesystem byte order "
+				"already normalized.\n"), ctx->device_name);
+			fatal_error(ctx, 0);
 		}
 	}
 	if (swapfs) {
 		swap_filesys(ctx);
 		if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
-			exit(FSCK_ERROR);
+			fatal_error(ctx, 0);
 	}
+#endif
 
 	/*
 	 * Mark the system as valid, 'til proven otherwise
@@ -771,26 +921,27 @@ restart:
 	retval = ext2fs_read_bb_inode(fs, &fs->badblocks);
 	if (retval) {
 		com_err(ctx->program_name, retval,
-			"while reading bad blocks inode");
+			_("while reading bad blocks inode"));
 		preenhalt(ctx);
-		printf("This doesn't bode well, but we'll try to go on...\n");
+		printf(_("This doesn't bode well,"
+			 " but we'll try to go on...\n"));
 	}
 
 	run_result = e2fsck_run(ctx);
 	e2fsck_clear_progbar(ctx);
 	if (run_result == E2F_FLAG_RESTART) {
-		printf("Restarting e2fsck from the beginning...\n");
+		printf(_("Restarting e2fsck from the beginning...\n"));
 		retval = e2fsck_reset_context(ctx);
 		if (retval) {
 			com_err(ctx->program_name, retval,
-				"while resetting context");
-			exit(1);
+				_("while resetting context"));
+			fatal_error(ctx, 0);
 		}
 		ext2fs_close(fs);
 		goto restart;
 	}
 	if (run_result & E2F_FLAG_SIGNAL_MASK)
-		exit(FSCK_ERROR);
+		fatal_error(ctx, 0);
 	if (run_result & E2F_FLAG_CANCEL)
 		ext2fs_unmark_valid(fs);
 
@@ -800,10 +951,10 @@ restart:
 	if (ext2fs_test_changed(fs)) {
 		exit_value = FSCK_NONDESTRUCT;
 		if (!(ctx->options & E2F_OPT_PREEN))
-			printf("\n%s: ***** FILE SYSTEM WAS MODIFIED *****\n",
+		    printf(_("\n%s: ***** FILE SYSTEM WAS MODIFIED *****\n"),
 			       ctx->device_name);
 		if (root_filesystem && !read_only_root) {
-			printf("%s: ***** REBOOT LINUX *****\n",
+			printf(_("%s: ***** REBOOT LINUX *****\n"),
 			       ctx->device_name);
 			exit_value = FSCK_REBOOT;
 		}
@@ -814,13 +965,13 @@ restart:
 		exit_value = FSCK_UNCORRECTED;
 	if (!(ctx->options & E2F_OPT_READONLY)) {
 		if (ext2fs_test_valid(fs)) {
-			if (!(fs->super->s_state & EXT2_VALID_FS))
+			if (!(sb->s_state & EXT2_VALID_FS))
 				exit_value = FSCK_NONDESTRUCT;
-			fs->super->s_state = EXT2_VALID_FS;
+			sb->s_state = EXT2_VALID_FS;
 		} else
-			fs->super->s_state &= ~EXT2_VALID_FS;
-		fs->super->s_mnt_count = 0;
-		fs->super->s_lastcheck = time(NULL);
+			sb->s_state &= ~EXT2_VALID_FS;
+		sb->s_mnt_count = 0;
+		sb->s_lastcheck = time(NULL);
 		ext2fs_mark_super_dirty(fs);
 	}
 	show_stats(ctx);
