@@ -39,6 +39,12 @@ void fatal_error(e2fsck_t ctx, const char *msg)
 {
 	if (msg) 
 		fprintf (stderr, "e2fsck: %s\n", msg);
+	if (ctx->fs && ctx->fs->io) {
+		if (ctx->fs->io->magic == EXT2_ET_MAGIC_IO_CHANNEL)
+			io_channel_flush(ctx->fs->io);
+		else
+			fprintf(stderr, "e2fsck: io manager magic bad!\n");
+	}
 	ctx->flags |= E2F_FLAG_ABORT;
 	if (ctx->flags & E2F_FLAG_SETJMP_OK)
 		longjmp(ctx->abort_loc, 1);
@@ -67,6 +73,8 @@ int ask_yn(const char * string, int def)
 {
 	int		c;
 	const char	*defstr;
+	const char	*short_yes = _("yY");
+	const char	*short_no = _("nN");
 
 #ifdef HAVE_TERMIOS_H
 	struct termios	termios, tmp;
@@ -80,22 +88,21 @@ int ask_yn(const char * string, int def)
 #endif
 
 	if (def == 1)
-		defstr = "<y>";
+		defstr = _("<y>");
 	else if (def == 0)
-		defstr = "<n>";
+		defstr = _("<n>");
 	else
-		defstr = " (y/n)";
+		defstr = _(" (y/n)");
 	printf("%s%s? ", string, defstr);
 	while (1) {
 		fflush (stdout);
 		if ((c = read_a_char()) == EOF)
 			break;
-		c = toupper(c);
-		if (c == 'Y') {
+		if (strchr(short_yes, (char) c)) {
 			def = 1;
 			break;
 		}
-		else if (c == 'N') {
+		else if (strchr(short_no, (char) c)) {
 			def = 0;
 			break;
 		}
@@ -115,15 +122,15 @@ int ask_yn(const char * string, int def)
 int ask (e2fsck_t ctx, const char * string, int def)
 {
 	if (ctx->options & E2F_OPT_NO) {
-		printf ("%s? no\n\n", string);
+		printf (_("%s? no\n\n"), string);
 		return 0;
 	}
 	if (ctx->options & E2F_OPT_YES) {
-		printf ("%s? yes\n\n", string);
+		printf (_("%s? yes\n\n"), string);
 		return 1;
 	}
 	if (ctx->options & E2F_OPT_PREEN) {
-		printf ("%s? %s\n\n", string, def ? "yes" : "no");
+		printf ("%s? %s\n\n", string, def ? _("yes") : _("no"));
 		return def;
 	}
 	return ask_yn(string, def);
@@ -136,17 +143,17 @@ void e2fsck_read_bitmaps(e2fsck_t ctx)
 
 	if (ctx->invalid_bitmaps) {
 		com_err(ctx->program_name, 0,
-			"e2fsck_read_bitmaps: illegal bitmap block(s) for %s",
+		    _("e2fsck_read_bitmaps: illegal bitmap block(s) for %s"),
 			ctx->device_name);
 		fatal_error(ctx, 0);
 	}
 
-	ehandler_operation("reading inode and block bitmaps");
+	ehandler_operation(_("reading inode and block bitmaps"));
 	retval = ext2fs_read_bitmaps(fs);
 	ehandler_operation(0);
 	if (retval) {
 		com_err(ctx->program_name, retval,
-			"while retrying to read bitmaps for %s",
+			_("while retrying to read bitmaps for %s"),
 			ctx->device_name);
 		fatal_error(ctx, 0);
 	}
@@ -158,24 +165,24 @@ void e2fsck_write_bitmaps(e2fsck_t ctx)
 	errcode_t	retval;
 
 	if (ext2fs_test_bb_dirty(fs)) {
-		ehandler_operation("writing block bitmaps");
+		ehandler_operation(_("writing block bitmaps"));
 		retval = ext2fs_write_block_bitmap(fs);
 		ehandler_operation(0);
 		if (retval) {
 			com_err(ctx->program_name, retval,
-				"while retrying to write block bitmaps for %s",
+			    _("while retrying to write block bitmaps for %s"),
 				ctx->device_name);
 			fatal_error(ctx, 0);
 		}
 	}
 
 	if (ext2fs_test_ib_dirty(fs)) {
-		ehandler_operation("writing inode bitmaps");
+		ehandler_operation(_("writing inode bitmaps"));
 		retval = ext2fs_write_inode_bitmap(fs);
 		ehandler_operation(0);
 		if (retval) {
 			com_err(ctx->program_name, retval,
-				"while retrying to write inode bitmaps for %s",
+			    _("while retrying to write inode bitmaps for %s"),
 				ctx->device_name);
 			fatal_error(ctx, 0);
 		}
@@ -188,8 +195,8 @@ void preenhalt(e2fsck_t ctx)
 
 	if (!(ctx->options & E2F_OPT_PREEN))
 		return;
-	fprintf(stderr, "\n\n%s: UNEXPECTED INCONSISTENCY; "
-		"RUN fsck MANUALLY.\n\t(i.e., without -a or -p options)\n",
+	fprintf(stderr, _("\n\n%s: UNEXPECTED INCONSISTENCY; "
+		"RUN fsck MANUALLY.\n\t(i.e., without -a or -p options)\n"),
 	       ctx->device_name);
 	if (fs != NULL) {
 		fs->super->s_state |= EXT2_ERROR_FS;
@@ -209,8 +216,8 @@ void init_resource_track(struct resource_track *track)
 	track->brk_start = sbrk(0);
 	gettimeofday(&track->time_start, 0);
 #ifdef HAVE_GETRUSAGE
-#ifdef solaris
-	memcpy(&r, 0, sizeof(struct rusage));
+#ifdef sun
+	memset(&r, 0, sizeof(struct rusage));
 #endif
 	getrusage(RUSAGE_SELF, &r);
 	track->user_start = r.ru_utime;
@@ -250,21 +257,25 @@ void print_resource_track(const char *desc, struct resource_track *track)
 		printf("%s: ", desc);
 
 #ifdef HAVE_MALLINFO
+#define kbytes(x)	(((x) + 1023) / 1024)
+	
 	malloc_info = mallinfo();
-	printf("Memory used: %d/%d, ", malloc_info.arena, malloc_info.hblkhd);
+	printf(_("Memory used: %dk/%dk (%dk/%dk), "),
+	       kbytes(malloc_info.arena), kbytes(malloc_info.hblkhd),
+	       kbytes(malloc_info.uordblks), kbytes(malloc_info.fordblks));
 #else
-	printf("Memory used: %d, ",
+	printf(_("Memory used: %d, "),
 	       (int) (((char *) sbrk(0)) - ((char *) track->brk_start)));
 #endif	
 #ifdef HAVE_GETRUSAGE
 	getrusage(RUSAGE_SELF, &r);
 
-	printf("elapsed time: %6.3f/%6.3f/%6.3f\n",
+	printf(_("time: %5.2f/%5.2f/%5.2f\n"),
 	       timeval_subtract(&time_end, &track->time_start),
 	       timeval_subtract(&r.ru_utime, &track->user_start),
 	       timeval_subtract(&r.ru_stime, &track->system_start));
 #else
-	printf("elapsed time: %6.3f\n",
+	printf(_("elapsed time: %6.3f\n"),
 	       timeval_subtract(&time_end, &track->time_start));
 #endif
 }
@@ -278,7 +289,7 @@ void e2fsck_read_inode(e2fsck_t ctx, unsigned long ino,
 	retval = ext2fs_read_inode(ctx->fs, ino, inode);
 	if (retval) {
 		com_err("ext2fs_read_inode", retval,
-			"while reading inode %ld in %s", ino, proc);
+			_("while reading inode %ld in %s"), ino, proc);
 		fatal_error(ctx, 0);
 	}
 }
@@ -291,7 +302,7 @@ extern void e2fsck_write_inode(e2fsck_t ctx, unsigned long ino,
 	retval = ext2fs_write_inode(ctx->fs, ino, inode);
 	if (retval) {
 		com_err("ext2fs_write_inode", retval,
-			"while writing inode %ld in %s", ino, proc);
+			_("while writing inode %ld in %s"), ino, proc);
 		fatal_error(ctx, 0);
 	}
 }
@@ -311,7 +322,7 @@ blk_t get_backup_sb(ext2_filsys fs)
 {
 	if (!fs || !fs->super)
 		return 8193;
-	return fs->super->s_blocks_per_group + 1;
+	return fs->super->s_blocks_per_group + fs->super->s_first_data_block;
 }
 
 /*
