@@ -14,6 +14,9 @@
  * 93/10/30	- Creation
  */
 
+#define _LARGEFILE_SOURCE
+#define _LARGEFILE64_SOURCE
+
 #if HAVE_ERRNO_H
 #include <errno.h>
 #endif
@@ -23,20 +26,24 @@
 #include <fcntl.h>
 #include <sys/ioctl.h>
 
-#include <linux/ext2_fs.h>
-
 #include "e2p.h"
 
+#ifdef O_LARGEFILE
+#define OPEN_FLAGS (O_RDONLY|O_NONBLOCK|O_LARGEFILE)
+#else
+#define OPEN_FLAGS (O_RDONLY|O_NONBLOCK)
+#endif
+
 int fgetversion (const char * name, unsigned long * version)
 {
 #if HAVE_EXT2_IOCTLS
-	int fd;
-	int r;
+	int fd, r, ver;
 
-	fd = open (name, O_RDONLY|O_NONBLOCK);
+	fd = open (name, OPEN_FLAGS);
 	if (fd == -1)
 		return - 1;
-	r = ioctl (fd, EXT2_IOC_GETVERSION, version);
+	r = ioctl (fd, EXT2_IOC_GETVERSION, &ver);
+	*version = ver;
 	close (fd);
 	return r;
 #else /* ! HAVE_EXT2_IOCTLS */
