@@ -23,13 +23,9 @@
 #include <sys/types.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
+#include "e2image.h"
 
 /*
  *  Note: if superblock is non-zero, block-size must also be non-zero.
@@ -39,7 +35,8 @@
  * 
  * 	EXT2_FLAG_RW	- Open the filesystem for read/write.
  * 	EXT2_FLAG_FORCE - Open the filesystem even if some of the
- *  				features aren't supported.
+ *				features aren't supported.
+ *	EXT2_FLAG_JOURNAL_DEV_OK - Open an ext3 journal device
  */
 errcode_t ext2fs_open(const char *name, int flags, int superblock,
 		      int block_size, io_manager manager, ext2_filsys *ret_fs)
@@ -50,7 +47,6 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 	blk_t		group_block;
 	char		*dest;
 	struct ext2_group_desc *gdp;
-	struct ext2fs_sb	*s;
 	
 	EXT2_CHECK_MAGIC(manager, EXT2_ET_MAGIC_IO_MANAGER);
 
@@ -74,12 +70,30 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 	retval = ext2fs_get_mem(SUPERBLOCK_SIZE, (void **) &fs->super);
 	if (retval)
 		goto cleanup;
+	if (flags & EXT2_FLAG_IMAGE_FILE) {
+		retval = ext2fs_get_mem(sizeof(struct ext2_image_hdr),
+					(void **) &fs->image_header);
+		if (retval)
+			goto cleanup;
+		retval = io_channel_read_blk(fs->io, 0,
+					     -sizeof(struct ext2_image_hdr),
+					     fs->image_header);
+		if (retval)
+			goto cleanup;
+		if (fs->image_header->magic_number != EXT2_ET_MAGIC_E2IMAGE)
+			return EXT2_ET_MAGIC_E2IMAGE;
+		superblock = 1;
+		block_size = fs->image_header->fs_blocksize;
+	}
 
 	/*
 	 * If the user specifies a specific block # for the
 	 * superblock, then he/she must also specify the block size!
 	 * Otherwise, read the master superblock located at offset
 	 * SUPERBLOCK_OFFSET from the start of the partition.
+	 *
+	 * Note: we only save a backup copy of the superblock if we
+	 * are reading the superblock from the primary superblock location.
 	 */
 	if (superblock) {
 		if (!block_size) {
@@ -88,55 +102,62 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 		}
 		io_channel_set_blksize(fs->io, block_size);
 		group_block = superblock + 1;
+		fs->orig_super = 0;
 	} else {
 		io_channel_set_blksize(fs->io, SUPERBLOCK_OFFSET);
 		superblock = 1;
 		group_block = 0;
+		retval = ext2fs_get_mem(SUPERBLOCK_SIZE,
+					(void **) &fs->orig_super);
+		if (retval)
+			goto cleanup;
 	}
 	retval = io_channel_read_blk(fs->io, superblock, -SUPERBLOCK_SIZE,
 				     fs->super);
 	if (retval)
 		goto cleanup;
+	if (fs->orig_super)
+		memcpy(fs->orig_super, fs->super, SUPERBLOCK_SIZE);
 
+#ifdef EXT2FS_ENABLE_SWAPFS
 	if ((fs->super->s_magic == ext2fs_swab16(EXT2_SUPER_MAGIC)) ||
 	    (fs->flags & EXT2_FLAG_SWAP_BYTES)) {
 		fs->flags |= EXT2_FLAG_SWAP_BYTES;
 
 		ext2fs_swap_super(fs->super);
 	}
+#endif
 	
 	if (fs->super->s_magic != EXT2_SUPER_MAGIC) {
 		retval = EXT2_ET_BAD_MAGIC;
 		goto cleanup;
 	}
-#ifdef EXT2_DYNAMIC_REV
-	if (fs->super->s_rev_level > EXT2_DYNAMIC_REV) {
-		retval = EXT2_ET_REV_TOO_HIGH;
-		goto cleanup;
-	}
-#else
-#ifdef	EXT2_CURRENT_REV
 	if (fs->super->s_rev_level > EXT2_LIB_CURRENT_REV) {
 		retval = EXT2_ET_REV_TOO_HIGH;
 		goto cleanup;
 	}
-#endif
-#endif
+
 	/*
 	 * Check for feature set incompatibility
 	 */
 	if (!(flags & EXT2_FLAG_FORCE)) {
-		s = (struct ext2fs_sb *) fs->super;
-		if (s->s_feature_incompat & ~EXT2_LIB_FEATURE_INCOMPAT_SUPP) {
+		if (fs->super->s_feature_incompat &
+		    ~EXT2_LIB_FEATURE_INCOMPAT_SUPP) {
 			retval = EXT2_ET_UNSUPP_FEATURE;
 			goto cleanup;
 		}
 		if ((flags & EXT2_FLAG_RW) &&
-		    (s->s_feature_ro_compat &
+		    (fs->super->s_feature_ro_compat &
 		     ~EXT2_LIB_FEATURE_RO_COMPAT_SUPP)) {
 			retval = EXT2_ET_RO_UNSUPP_FEATURE;
 			goto cleanup;
 		}
+		if (!(flags & EXT2_FLAG_JOURNAL_DEV_OK) &&
+		    (fs->super->s_feature_incompat &
+		     EXT3_FEATURE_INCOMPAT_JOURNAL_DEV)) {
+			retval = EXT2_ET_UNSUPP_FEATURE;
+			goto cleanup;
+		}
 	}
 	
 	fs->blocksize = EXT2_BLOCK_SIZE(fs->super);
@@ -159,6 +180,17 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 	 * Set the blocksize to the filesystem's blocksize.
 	 */
 	io_channel_set_blksize(fs->io, fs->blocksize);
+
+	/*
+	 * If this is an external journal device, don't try to read
+	 * the group descriptors, because they're not there.
+	 */
+	if (fs->super->s_feature_incompat &
+	    EXT3_FEATURE_INCOMPAT_JOURNAL_DEV) {
+		fs->group_desc_count = 0;
+		*ret_fs = fs;
+		return 0;
+	}
 	
 	/*
 	 * Read group descriptors
@@ -186,6 +218,7 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 		if (retval)
 			goto cleanup;
 		group_block++;
+#ifdef EXT2FS_ENABLE_SWAPFS
 		if (fs->flags & EXT2_FLAG_SWAP_BYTES) {
 			gdp = (struct ext2_group_desc *) dest;
 			groups_per_block = fs->blocksize /
@@ -193,6 +226,7 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 			for (j=0; j < groups_per_block; j++)
 				ext2fs_swap_group_desc(gdp++);
 		}
+#endif
 		dest += fs->blocksize;
 	}
 
