@@ -1,7 +1,7 @@
 /*
- * read_bb_file.c --- read a list of bad blocks for a FILE *
+ * read_bb_file.c --- read a list of bad blocks from a FILE *
  *
- * Copyright (C) 1994, 1995 Theodore Ts'o.
+ * Copyright (C) 1994, 1995, 2000 Theodore Ts'o.
  *
  * %Begin-Header%
  * This file may be redistributed under the terms of the GNU Public
@@ -23,27 +23,27 @@
 #include <sys/types.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
 
 /*
  * Reads a list of bad blocks from  a FILE *
  */
-errcode_t ext2fs_read_bb_FILE(ext2_filsys fs, FILE *f, 
-			      ext2_badblocks_list *bb_list,
-			      void (*invalid)(ext2_filsys fs, blk_t blk))
+errcode_t ext2fs_read_bb_FILE2(ext2_filsys fs, FILE *f, 
+			       ext2_badblocks_list *bb_list,
+			       void *private,
+			       void (*invalid)(ext2_filsys fs,
+					       blk_t blk,
+					       char *badstr,
+					       void *private))
 {
 	errcode_t	retval;
 	blk_t		blockno;
 	int		count;
 	char		buf[128];
 
-	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
+	if (fs)
+		EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
 
 	if (!*bb_list) {
 		retval = ext2fs_badblocks_list_create(bb_list, 10);
@@ -57,10 +57,11 @@ errcode_t ext2fs_read_bb_FILE(ext2_filsys fs, FILE *f,
 		count = sscanf(buf, "%u", &blockno);
 		if (count <= 0)
 			continue;
-		if ((blockno < fs->super->s_first_data_block) ||
-		    (blockno >= fs->super->s_blocks_count)) {
+		if (fs &&
+		    ((blockno < fs->super->s_first_data_block) ||
+		    (blockno >= fs->super->s_blocks_count))) {
 			if (invalid)
-				(invalid)(fs, blockno);
+				(invalid)(fs, blockno, buf, private);
 			continue;
 		}
 		retval = ext2fs_badblocks_list_add(*bb_list, blockno);
@@ -70,4 +71,26 @@ errcode_t ext2fs_read_bb_FILE(ext2_filsys fs, FILE *f,
 	return 0;
 }
 
+static void call_compat_invalid(ext2_filsys fs, blk_t blk,
+				char *badstr, void *private)
+{
+	void (*invalid)(ext2_filsys, blk_t);
+
+	invalid = (void (*)(ext2_filsys, blk_t)) private;
+	if (invalid)
+		invalid(fs, blk);
+}
+
+
+/*
+ * Reads a list of bad blocks from  a FILE *
+ */
+errcode_t ext2fs_read_bb_FILE(ext2_filsys fs, FILE *f, 
+			      ext2_badblocks_list *bb_list,
+			      void (*invalid)(ext2_filsys fs, blk_t blk))
+{
+	return ext2fs_read_bb_FILE2(fs, f, bb_list, (void *) invalid,
+				    call_compat_invalid);
+}
+
 
