@@ -1,8 +1,8 @@
 /*
  * Listener loop for subsystem library libss.a.
  *
- *	$Header: /usr/src/CVS-REPO/e2fsprogs/lib/ss/listen.c,v 1.15 1999/10/23 01:16:22 tytso Exp $
- *	$Locker:  $
+ *	$Header: /usr/src/e2fsprogs/e2fsprogs/lib/ss/SCCS/s.listen.c 1.15 99/10/23 01:16:22-00:00 tytso@mit.edu $
+ *	$Locker: <Not implemented> $
  * 
  * Copyright 1987, 1988 by MIT Student Information Processing Board
  *
@@ -27,7 +27,7 @@
 
 #ifndef	lint
 static char const rcs_id[] =
-    "$Header: /usr/src/CVS-REPO/e2fsprogs/lib/ss/listen.c,v 1.15 1999/10/23 01:16:22 tytso Exp $";
+    "$Header: /usr/src/e2fsprogs/e2fsprogs/lib/ss/SCCS/s.listen.c 1.15 99/10/23 01:16:22-00:00 tytso@mit.edu $";
 #endif
 
 typedef void sigret_t;
