@@ -4,13 +4,13 @@
  * Copyright (C) 1996, 1997 Theodore Ts'o.
  *
  * %Begin-Header%
- * This file may be redistributed under the terms of the GNU Public
- * License.
+ * This file may be redistributed under the terms of the GNU 
+ * Library General Public License.
  * %End-Header%
  */
 
 #include <sys/types.h>
-#include <asm/types.h>
+#include <ext2fs/ext2_types.h>
 
 #include "uuid.h"
 
