@@ -6,8 +6,8 @@
  * Copyright (C) 1998, 1999 Theodore Ts'o.
  *
  * %Begin-Header%
- * This file may be redistributed under the terms of the GNU Public
- * License.
+ * This file may be redistributed under the terms of the GNU 
+ * Library General Public License.
  * %End-Header%
  */
 
@@ -17,7 +17,6 @@
 #include <sys/types.h>
 #include <sys/time.h>
 #include <time.h>
-#include <linux/ext2_fs.h>
 
 #include "uuidP.h"
 
@@ -131,7 +130,7 @@ main(int argc, char **argv)
 		printf("Warning: not a time-based UUID, so UUID time "
 		       "decoding will likely not work!\n");
 	}
-	printf("UUID time is: (%d, %d): %s\n", tv.tv_sec, tv.tv_usec,
+	printf("UUID time is: (%ld, %ld): %s\n", tv.tv_sec, tv.tv_usec,
 	       ctime(&time_reg));
 	
 	return 0;
