@@ -1,4 +1,27 @@
-/* Well, here's my linux version of findsuper.
+/*
+ * findsuper --- quick hacked up program to find ext2 superblocks.
+ *
+ * This is a hack, and really shouldn't be installed anywhere.  If you
+ * need a program which does this sort of functionality, please try
+ * using gpart program.
+ *
+ * Portions Copyright 1998-2000, Theodore Ts'o.
+ * 
+ * This program may be used under the provisions of the GNU Public
+ * License, *EXCEPT* that a binary copy of the executable may not be
+ * packaged as a part of binary package which is distributed as part
+ * of a Linux distribution.  (Yes, this violates the Debian Free
+ * Software Guidelines in terms of restricting its field of use.
+ * That's the point.  I don't want this program being distributed in
+ * Debian, because I don't care to support it, and the maintainer,
+ * Yann Dirson, doesn't seem to pay attention to my wishes on this
+ * matter.  So I'm delibiately adding this clause so it violates the
+ * Debian Free Software Guidelines to force him to take it out.  If
+ * this doesn't work, I'll have to remove it from the upstream source
+ * distribution at the next release.  End of Rant.  :-)
+ * 
+ *
+ * Well, here's my linux version of findsuper.
  * I'm sure you coulda done it faster.  :)
  * IMHO there isn't as much interesting data to print in the
  * linux superblock as there is in the SunOS superblock--disk geometry is
@@ -67,7 +90,8 @@
 #include <stdlib.h>
 #include <time.h>
 
-#include <linux/ext2_fs.h>
+#include "ext2fs/ext2_fs.h"
+#include "nls-enable.h"
 
 
 main(int argc, char *argv[])
@@ -82,23 +106,28 @@ main(int argc, char *argv[])
 	struct ext2_super_block ext2;
 	/* interesting fields: EXT2_SUPER_MAGIC
 	 *      s_blocks_count s_log_block_size s_mtime s_magic s_lastcheck */
-  
+
+#ifdef ENABLE_NLS
+	setlocale(LC_MESSAGES, "");
+	bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
+	textdomain(NLS_CAT_NAME);
+#endif
 	if (argc<2) {
 		fprintf(stderr,
-			"Usage:  findsuper device [skiprate [start]]\n");
+			_("Usage:  findsuper device [skiprate [start]]\n"));
 		exit(1);
 	}
 	if (argc>2)
 		skiprate=atoi(argv[2]);
 	if (skiprate<512) {
 		fprintf(stderr,
-			"Do you really want to skip less than a sector??\n");
+			_("Do you really want to skip less than a sector??\n"));
 		exit(2);
 	}
 	if (argc>3)
 		sk=atol(argv[3]);
 	if (sk<0) {
-		fprintf(stderr,"Have to start at 0 or greater,not %ld\n",sk);
+		fprintf(stderr,_("Have to start at 0 or greater,not %ld\n"),sk);
 		exit(1);
 	}
 	f=fopen(argv[1],"r");
@@ -111,7 +140,7 @@ main(int argc, char *argv[])
 	printf("  thisoff     block fs_blk_sz  blksz grp last_mount\n");
 	for (;!feof(f) &&  (i=fseek(f,sk,SEEK_SET))!= -1; sk+=skiprate){
 		if (i=fread(&ext2,sizeof(ext2),1, f)!=1) {
-			perror("read failed");
+			perror(_("read failed"));
 		}
 		if (ext2.s_magic != EXT2_SUPER_MAGIC)
 			continue;
@@ -124,6 +153,6 @@ main(int argc, char *argv[])
 		       ext2.s_log_block_size,
 		       ext2.s_block_group_nr, s);
 	}
-	printf("Failed on %d at %ld\n", i, sk);
+	printf(_("Failed on %d at %ld\n"), i, sk);
 	fclose(f);
 }
