@@ -2,6 +2,8 @@
  * fsck.h
  */
 
+#include <time.h>
+
 #ifdef __STDC__
 #define NOARGS void
 #else
@@ -53,8 +55,9 @@ struct fsck_instance {
 	char *	prog;
 	char *	type;
 	char *	device;
+	char *	base_device;
 	struct fsck_instance *next;
 };
 
-
-
+extern char *base_device(char *device);
+extern char *string_copy(const char *s);
