@@ -17,6 +17,8 @@
  * 98/12/29	- Display version info only when -V specified (G M Sipe)
  */
 
+#define _LARGEFILE64_SOURCE
+
 #include <sys/types.h>
 #include <dirent.h>
 #ifdef HAVE_ERRNO_H
@@ -35,45 +37,63 @@ extern char *optarg;
 #include <string.h>
 #include <sys/param.h>
 #include <sys/stat.h>
-#include <linux/ext2_fs.h>
 
+#include "ext2fs/ext2_fs.h"
 #include "et/com_err.h"
 #include "e2p/e2p.h"
 
 #include "../version.h"
+#include "nls-enable.h"
 
-const char * program_name = "lsattr";
+static const char * program_name = "lsattr";
 
-int all = 0;
-int d_opt = 0;
-int l_opt = 0;
-int recursive = 0;
-int verbose = 0;
-int v_opt = 0;
+static int all;
+static int dirs_opt;
+static unsigned pf_options;
+static int recursive;
+static int verbose;
+static int generation_opt;
+
+#ifdef _LFS64_LARGEFILE
+#define LSTAT		lstat64
+#define STRUCT_STAT	struct stat64
+#else
+#define LSTAT		lstat
+#define STRUCT_STAT	struct stat
+#endif
 
 static void usage(void)
 {
-	fprintf(stderr, "Usage: %s [-RVadlv] [files...]\n", program_name);
+	fprintf(stderr, _("Usage: %s [-RVadlv] [files...]\n"), program_name);
 	exit(1);
 }
 
 static void list_attributes (const char * name)
 {
 	unsigned long flags;
-	unsigned long version;
+	unsigned long generation;
 
-	if (fgetflags (name, &flags) == -1)
-		com_err (program_name, errno, "While reading flags on %s",
+	if (fgetflags (name, &flags) == -1) {
+		com_err (program_name, errno, _("While reading flags on %s"),
 			 name);
-	else if (fgetversion (name, &version) == -1)
-		com_err (program_name, errno, "While reading version on %s",
-			 name);
-	else
-	{
-		if (v_opt)
-			printf ("%5lu ", version);
-		print_flags (stdout, flags, l_opt);
-		printf (" %s\n", name);
+		return;
+	}
+	if (generation_opt) {
+		if (fgetversion (name, &generation) == -1) {
+			com_err (program_name, errno,
+				 _("While reading version on %s"),
+				 name);
+			return;
+		}
+		printf ("%5lu ", generation);
+	}
+	if (pf_options & PFOPT_LONG) {
+		printf("%-28s ", name);
+		print_flags(stdout, flags, pf_options);
+		fputc('\n', stdout);
+	} else {
+		print_flags(stdout, flags, pf_options);
+		printf(" %s\n", name);
 	}
 }
 
@@ -81,14 +101,14 @@ static int lsattr_dir_proc (const char *, struct dirent *, void *);
 
 static void lsattr_args (const char * name)
 {
-	struct stat st;
+	STRUCT_STAT	st;
 
-	if (lstat (name, &st) == -1)
-		com_err (program_name, errno, "while stating %s", name);
-	else
-	{
-		if (S_ISDIR(st.st_mode) && !d_opt)
-			iterate_on_dir (name, lsattr_dir_proc, (void *) NULL);
+	if (LSTAT (name, &st) == -1)
+		com_err (program_name, errno, _("while trying to stat %s"),
+			 name);
+	else {
+		if (S_ISDIR(st.st_mode) && !dirs_opt)
+			iterate_on_dir (name, lsattr_dir_proc, NULL);
 		else
 			list_attributes (name);
 	}
@@ -96,13 +116,13 @@ static void lsattr_args (const char * name)
 
 static int lsattr_dir_proc (const char * dir_name, struct dirent * de, void * private)
 {
-	struct stat st;
+	STRUCT_STAT	st;
 	char *path;
 
 	path = malloc(strlen (dir_name) + 1 + strlen (de->d_name) + 1);
 
 	sprintf (path, "%s/%s", dir_name, de->d_name);
-	if (lstat (path, &st) == -1)
+	if (LSTAT (path, &st) == -1)
 		perror (path);
 	else {
 		if (de->d_name[0] != '.' || all) {
@@ -111,8 +131,7 @@ static int lsattr_dir_proc (const char * dir_name, struct dirent * de, void * pr
 			    strcmp(de->d_name, ".") &&
 			    strcmp(de->d_name, "..")) {
 				printf ("\n%s:\n", path);
-				iterate_on_dir (path, lsattr_dir_proc,
-						(void *) NULL);
+				iterate_on_dir (path, lsattr_dir_proc, NULL);
 				printf ("\n");
 			}
 		}
@@ -126,6 +145,11 @@ int main (int argc, char ** argv)
 	int c;
 	int i;
 
+#ifdef ENABLE_NLS
+	setlocale(LC_MESSAGES, "");
+	bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
+	textdomain(NLS_CAT_NAME);
+#endif
 	if (argc && *argv)
 		program_name = *argv;
 	while ((c = getopt (argc, argv, "RVadlv")) != EOF)
@@ -141,20 +165,20 @@ int main (int argc, char ** argv)
 				all = 1;
 				break;
 			case 'd':
-				d_opt = 1;
+				dirs_opt = 1;
 				break;
 			case 'l':
-				l_opt = 1;
+				pf_options = PFOPT_LONG;
 				break;
 			case 'v':
-				v_opt = 1;
+				generation_opt = 1;
 				break;
 			default:
 				usage();
 		}
 
 	if (verbose)
-		fprintf (stderr, "lsattr %s, %s for EXT2 FS %s, %s\n",
+		fprintf (stderr, _("lsattr %s, %s for EXT2 FS %s, %s\n"),
 			 E2FSPROGS_VERSION, E2FSPROGS_DATE,
 			 EXT2FS_VERSION, EXT2FS_DATE);
 	if (optind > argc - 1)
